from visidata import vd, VisiData, ItemColumn, AttrDict


from .gitsheet import GitSheet
from .diff import GitDiffSheet


@VisiData.api
def git_stash(vd, p, args):
    if 'list' in args:
        return GitStashes('git-stash-list', source=p, gitargs=args)


class GitStashes(GitSheet):
    guide = '''
        # git stash
        This is the list of changes that have been stashed previously.

        `a` to apply this stashed change (without removing it)
        `d` to drop this stashed change
        `b` to create a branch from this stashed change'),
    '''
    rowtype = 'stashed change'  # rowdef: AttrDict(stashid=, branched_from=, sha1=, msg=)
    columns = [
        ItemColumn('stashid'),
        ItemColumn('branched_from'),
        ItemColumn('sha1'),
        ItemColumn('msg'),
        ItemColumn('line', width=0),
    ]

    def iterload(self):
        for line in self.git_lines('stash', *self.gitargs):
            stashid, ctx, rest = line.split(': ', 2)
            if ctx.startswith('WIP on '):
                branched_from = ctx[len('WIP on '):]
                sha1, msg = rest.split(' ', 1)
            elif ctx.startswith('On '):
                branched_from = ctx[len('On '):]
                sha1 = ''
                msg = rest
            yield AttrDict(
                line=line,
                stashid=stashid,
                branched_from=branched_from,
                sha1=sha1,
                msg=msg.strip(),
            )

    def openRow(self, row):
        'open this stashed change'
        return GitDiffSheet(row.stashid, "diffs", gitargs=['stash show --no-color --patch', row.stashid], source=self.source)


GitSheet.addCommand('', 'git-open-stashes', 'vd.push(git_stash(source, ["list"]))', 'push stashes sheet')

GitStashes.addCommand('a', 'git-stash-apply', 'loggit("stash", "apply", cursorRow[0])', 'apply this stashed change without removing')
GitStashes.addCommand('', 'git-stash-pop', 'loggit("stash", "pop", cursorRow[0])', 'apply this stashed change and drop it')
GitStashes.addCommand('d', 'git-stash-drop', 'loggit("stash", "drop", cursorRow[0])', 'drop this stashed change')
GitStashes.addCommand('b', 'git-stash-branch', 'loggit("stash", "branch", input("create branch from stash named: "), cursorRow[0])', 'create branch from stash')


vd.addMenuItems('''
    Git > Open > stashes > git-open-stashes
    Git > Stash > apply > git-stash-apply
    Git > Stash > drop > git-stash-drop
    Git > Stash > apply then drop  > git-stash-pop
    Git > Stash > create branch > git-stash-branch
''')



import os.path
import os
import sys
import re
import shutil
import importlib
import subprocess
import urllib.error

from visidata import VisiData, vd, Path, CellColorizer, Sheet, AttrDict, ItemColumn, Column, Progress, ExpectedException, BaseSheet, asyncthread


vd.option('plugins_autoload', True, 'do not autoload plugins if False')


@VisiData.property
def pluginConfig(self):
    return Path(os.path.join(vd.options.visidata_dir, "plugins", "__init__.py"))


@VisiData.property
def pluginConfigLines(self):
    return Path(self.pluginConfig).open(mode='r', encoding='utf-8').readlines()

def _plugin_import_name(self, plugin):
    if not plugin.url:
        return 'visidata.plugins.'+plugin.name
    if 'git+' in plugin.url:
        return plugin.name
    return "plugins."+plugin.name


@VisiData.api
def enablePlugin(vd, plugin:str):
    with vd.pluginConfig.open(mode='a', encoding='utf-8') as fprc:
        print(f'import {plugin}', file=fprc)
        importlib.import_module(plugin)
        vd.status(f'{plugin} plugin enabled')

@VisiData.api
def removePlugin(vd, plugin:str):
    path = vd.pluginConfig
    pathbackup = path.with_suffix(path.suffix + '.bak')
    try:
        shutil.copyfile(path, pathbackup)

        # Copy lines from the backup init file into its replacement, skipping lines that import the removed plugin.
        #
        # By matching from the start of a line through a word boundary, we avoid removing commented lines or inadvertently removing
        # plugins with similar names.

        r = re.compile(f'^import {plugin}\\W')
        nonimports = [line for line in vd.pluginConfigLines if not r.match(line)]
        if len(nonimports) == len(vd.pluginConfigLines):
            vd.fail("plugin not in import list")

        with path.open(mode='w', encoding='utf-8') as new:
            new.writelines(nonimports)

        sys.modules.pop(plugin)
        importlib.invalidate_caches()
        vd.warning(f'"{plugin}" plugin removed')
    except FileNotFoundError:
        vd.debug("no {vd.pluginConfig} found")


@VisiData.lazy_property
def pluginsSheet(p):
    return PluginsSheet('plugins')


class PluginsSheet(Sheet):
    rowtype = "plugins"  # rowdef: AttrDict of json dict
    colorizers = [
            CellColorizer(3, 'color_working', lambda s,c,r,v: r and r.installed)
    ]
    columns = [
        ItemColumn('name'),
        ItemColumn('installed', width=8),
        ItemColumn('description', width=60),
    ]
    nKeys = 1
    def iterload(self):
        import pkgutil
        import ast
        # enumerate installed plugins
        for name, mod in sys.modules.items():
            if name.startswith(('visidata.plugins.', 'visidata.experimental.')):
                yield AttrDict(name=name, # '.'.join(name.split('.')[2:]),
                               description=getattr(mod, '__description__', mod.__doc__),
                               installed=getattr(mod, '__version__', 'yes'),
                               maintainer=getattr(mod, '__author__', None))


BaseSheet.addCommand(None, 'open-plugins', 'vd.push(vd.pluginsSheet)', 'Open Plugins Sheet to manage supported plugins')

PluginsSheet.addCommand('a', 'add-plugin', 'enablePlugin(cursorRow.name); reload_rows()', 'Enable current plugin by adding to imports')
PluginsSheet.addCommand('d', 'delete-plugin', 'removePlugin(cursorRow.name); reload_rows()', 'Disable current plugin by removing from imports')

vd.addMenuItems('''
    System > Plugins Sheet > open-plugins
''')



import pathlib
import tarfile
import zipfile
import datetime
import os.path
from visidata.loaders import unzip_http

from visidata import vd, VisiData, asyncthread, Sheet, Progress, Menu, options
from visidata import ColumnAttr, Column, Path, filesize
from visidata.type_date import date

@VisiData.api
def guess_zip(vd, p):
    if not p.is_url() and zipfile.is_zipfile(p.open_bytes()):
        return dict(filetype='zip', _likelihood=10)

@VisiData.api
def guess_tar(vd, p):
    # an empty file will pass is_tarfile(), but can't be opened by tarfile.open()
    if filesize(p) == 0:
        return None
    if tarfile.is_tarfile(p.open_bytes()):
        return dict(filetype='tar', _likelihood=10)

@VisiData.api
def open_zip(vd, p):
    return vd.ZipSheet(p.base_stem, source=p)

@VisiData.api
def open_tar(vd, p):
    return TarSheet(p.base_stem, source=p)

VisiData.open_tgz = VisiData.open_tar
VisiData.open_txz = VisiData.open_tar
VisiData.open_tbz2 = VisiData.open_tar
VisiData.open_whl = VisiData.open_zip

@VisiData.api
class ZipSheet(Sheet):
    'Wrapper for `zipfile` library.'
    rowtype = 'files' # rowdef [ZipInfo, zipfile.Path]
    columns = [
        Column('directory',
            getter=lambda col,row: str(row[1].parent) if str(row[1].parent) == '.' else str(row[1].parent) + '/'),
        Column('filename', getter=lambda col,row: row[1].name + row[1].suffix),
        Column('abspath', type=str, width=0, getter=lambda col,row: row[1]),
        Column('ext', getter=lambda col,row: row[0].filename.endswith('/') and '/' or row[1].ext),
        Column('size', getter=lambda col,row: row[0].file_size, type=int),
        Column('compressed_size', type=int, getter=lambda col,row: row[0].compress_size),
        Column('date_time', type=date,
               getter=lambda col, row: datetime.datetime(*row[0].date_time)),
    ]
    nKeys = 2
    guide = '''# Zip Sheet
This is a list of files contained in the zipfile {sheet.displaySource}.

Once extracted, files can be loaded with `ENTER`.

Commands:

- `x` to extract current file to current directory
- `gx` to extract selected files to current directory
- `zx` to extract current file to a given pathname
- `gzx`  to extract selected files to given directory

'''

    def openZipFile(self, fp, *args, **kwargs):
        '''Use VisiData input to handle password-protected zip files.'''
        if isinstance(fp, zipfile.ZipFile):
            zip_open =  fp.open
        elif isinstance(fp, unzip_http.RemoteZipFile):
            zip_open = fp._open
        try:
            return zip_open(*args, **kwargs)
        except RuntimeError as err:
            if 'password required' in err.args[0]:
                pwd = vd.input(f'{args[0].filename} is encrypted, enter password: ', display=False)
                return zip_open(*args, **kwargs, pwd=pwd.encode('utf-8'))
            vd.exceptionCaught(err)

    def openRow(self, row):
            fi, zpath = row
            fp = self.openZipFile(self.zfp, fi)
            return vd.openSource(Path(fi.filename, fp=fp, filesize=fi.file_size), filetype=options.filetype)

    def extract(self, *rows, path=None):
        path = path or Path('.')

        files = []
        for row in rows:
            r, _ = row
            vd.confirmOverwrite(path/r.filename)  #1452
            self.extract_async(row, path=path)

    def sysopen_row(self, row):
        'Extract file in row to tempdir and launch $EDITOR.  Modifications will be discarded.'
        import tempfile
        with tempfile.TemporaryDirectory() as tempdir:
            self.zfp.extract(member=row[0], path=tempdir)
            vd.launchExternalEditorPath(Path(tempdir)/row[0].filename)

    @asyncthread
    def extract_async(self, *rows, path=None):
        'Extract rows to *path*, without confirmation.'
        for r, _ in Progress(rows):
            self.zfp.extract(member=r.filename, path=path)
            vd.status(f'extracted {r.filename}')

    @property
    def zfp(self):
        if not self._zfp:
            if '://' in str(self.source):
                unzip_http.warning = vd.warning
                self._zfp = unzip_http.RemoteZipFile(str(self.source))
            elif isinstance(self.source, Path):
                if self.source.has_fp():  #when opening a zip inside tar or zip
                    fp = self.source.open('rb')
                else:
                    fp = self.source
                self._zfp = zipfile.ZipFile(fp, 'r')
            else:
                self._zfp = zipfile.ZipFile(str(self.source), 'r')

        return self._zfp

    def iterload(self):
        for zi in Progress(self.zfp.infolist()):
            yield [zi, Path(zi.filename)]


#from https://docs.python.org/3/library/tarfile.html#tarfile.REGTYPE
tarfile_type_names = {
    tarfile.REGTYPE:"file",
    tarfile.AREGTYPE:"file",
    tarfile.LNKTYPE:"hard link",
    tarfile.SYMTYPE:"symbolic link",
    tarfile.CHRTYPE:"character device",
    tarfile.BLKTYPE:"block device",
    tarfile.DIRTYPE:"directory",
    tarfile.FIFOTYPE:"FIFO",
    tarfile.CONTTYPE:"contiguous file",
    tarfile.GNUTYPE_LONGNAME:"GNU tar longname",
    tarfile.GNUTYPE_LONGLINK:"GNU tar longlink",
    tarfile.GNUTYPE_SPARSE:"GNU tar sparse file",
}
class TarSheet(Sheet):
    'Wrapper for `tarfile` library.'
    rowtype = 'files' # rowdef TarInfo
    columns = [
        ColumnAttr('name'),
        Column('ext', getter=lambda col,row: row.isdir() and '/' or os.path.splitext(row.name)[1][1:]),
        ColumnAttr('size', type=int),
        ColumnAttr('mtime', type=date),
        Column('type', getter=lambda col, row: tarfile_type_names.get(row.type, 'unknown')),
        ColumnAttr('mode', type=int),
        ColumnAttr('uname'),
        ColumnAttr('gname')
    ]
    nKeys=1

    def openRow(self, fi):
            tfp = tarfile.open(name=str(self.source))
            return vd.openSource(Path(fi.name, fp=tfp.extractfile(fi), filesize=fi.size))

    def iterload(self):
        with tarfile.open(name=str(self.source)) as tf:
            for ti in Progress(tf.getmembers()):
                yield ti


ZipSheet.init('_zfp', lambda: None, copy=True)

ZipSheet.addCommand('x', 'extract-file', 'extract(cursorRow)', 'extract current file to current directory')
ZipSheet.addCommand('gx', 'extract-selected', 'extract(*onlySelectedRows)', 'extract selected files to current directory')
ZipSheet.addCommand('zx', 'extract-file-to', 'extract(cursorRow, path=inputPath("extract to: "))', 'extract current file to given pathname')
ZipSheet.addCommand('gzx', 'extract-selected-to', 'extract(*onlySelectedRows, path=inputPath("extract %d files to: " % nSelectedRows))', 'extract selected files to given directory')
ZipSheet.addCommand('Ctrl+O', 'sysopen-row', 'sysopen_row(cursorRow)', 'open $EDITOR with current file (modifications will be discarded)')

vd.addMenu(Menu('File', Menu('Extract',
        Menu('current file', 'extract-file'),
        Menu('current file to', 'extract-file-to'),
        Menu('selected files', 'extract-selected'),
        Menu('selected files to', 'extract-selected-to'),
    )))

vd.addGlobals({
    'ZipSheet': ZipSheet,
    'TarSheet': TarSheet
})



import unicodedata
import sys
import re
import functools
import textwrap

from visidata import options, drawcache, vd, update_attr, colors, ColorAttr

disp_column_fill = ' '
internal_markup_re = r'(\[[:/][^\]]*?\])'  # [:whatever until the closing bracket] or [/whatever] or [:]

### Curses helpers

# ZERO_WIDTH_CF is from wcwidth:
# NOTE: created by hand, there isn't anything identifiable other than
# general Cf category code to identify these, and some characters in Cf
# category code are of non-zero width.
# Also includes some Cc, Mn, Zl, and Zp characters
ZERO_WIDTH_CF = set(map(chr, [
    0,       # Null (Cc)
    0x034F,  # Combining grapheme joiner (Mn)
    0x200B,  # Zero width space
    0x200C,  # Zero width non-joiner
    0x200D,  # Zero width joiner
    0x200E,  # Left-to-right mark
    0x200F,  # Right-to-left mark
    0x2028,  # Line separator (Zl)
    0x2029,  # Paragraph separator (Zp)
    0x202A,  # Left-to-right embedding
    0x202B,  # Right-to-left embedding
    0x202C,  # Pop directional formatting
    0x202D,  # Left-to-right override
    0x202E,  # Right-to-left override
    0x2060,  # Word joiner
    0x2061,  # Function application
    0x2062,  # Invisible times
    0x2063,  # Invisible separator
]))

def wcwidth(cc, ambig=1):
        if cc in ZERO_WIDTH_CF:
            return 1
        eaw = unicodedata.east_asian_width(cc)
        if eaw in 'AN':  # ambiguous or neutral
            if unicodedata.category(cc) == 'Mn':
                return 1
            else:
                return ambig
        elif eaw in 'WF': # wide/full
            return 2
        elif not unicodedata.combining(cc):
            return 1
        return 0


def is_vdcode(s:str) -> bool:
    return (s.startswith('[:') and s.endswith(']')) or \
           (s.startswith('[/') and s.endswith(']'))


def iterchunks(s, literal=False):
    attrstack = [dict(link='', cattr=ColorAttr())]
    legitopens = 0
    chunks = re.split(internal_markup_re, s)
    for chunk in chunks:
        if not chunk:
            continue

        if not literal and is_vdcode(chunk):
            cattr = attrstack[-1]['cattr']
            link = attrstack[-1]['link']

            if chunk.startswith('[:onclick '):
                attrstack.append(dict(link=chunk[2:-1], cattr=cattr.update(colors.clickable)))
                continue
            elif chunk == '[:]':  # clear stack, keep origattr
                if len(attrstack) > 1:
                    del attrstack[1:]
                    continue
            elif chunk.startswith('[/'):  # pop last attr off stack
                if len(attrstack) > 1:
                    attrstack.pop()
                continue  # don't display trailing [/foo] ever
            else:  # push updated color on stack
                newcolor = colors.get_color(chunk[2:-1])
                if newcolor:
                    cattr = update_attr(cattr, newcolor, len(attrstack))
                    attrstack.append(dict(link=link, cattr=cattr))
                    continue

        yield attrstack[-1], chunk


@functools.lru_cache(maxsize=100000)
def dispwidth(ss, maxwidth=None, literal=False):
    'Return display width of string, according to unicodedata width and options.disp_ambig_width.'
    disp_ambig_width = options.disp_ambig_width
    w = 0

    for _, s in iterchunks(ss, literal=literal):
        for cc in s:
            if cc:
                w += wcwidth(cc, disp_ambig_width)
                if maxwidth and w > maxwidth:
                    return maxwidth
    return w


@functools.lru_cache(maxsize=100000)
def _dispch(c, oddspacech=None, combch=None, modch=None):
    ccat = unicodedata.category(c)
    if ccat in ['Mn', 'Sk', 'Lm']:
        if unicodedata.name(c).startswith('MODIFIER'):
            return modch, 1
    elif c != ' ' and ccat in ('Cc', 'Zs', 'Zl', 'Cs'):  # control char, space, line sep, surrogate
        return oddspacech, 1
    elif c in ZERO_WIDTH_CF:
        return combch, 1

    return c, dispwidth(c, literal=True)


def iterchars(x):
    if isinstance(x, dict):
        yield from '{%d}' % len(x)
        for k, v in x.items():
            yield ' '
            yield from iterchars(k)
            yield '='
            yield from iterchars(v)

    elif isinstance(x, (list, tuple)):
        yield from '[%d] ' % len(x)
        for i, v in enumerate(x):
            if i != 0:
                yield from '; '
            yield from iterchars(v)

    else:
        yield from str(x)


@functools.lru_cache(maxsize=100000)
def _clipstr(s, dispw, trunch='', oddspacech='', combch='', modch=''):
    ''' *s* is a string or an iterator that contains characters.
    *dispw* is the integer screen width that the clipped string will fit inside, or None.
    Return clipped string and width in terminal display characters.
    Note: width may differ from len(s) if chars are 'fullwidth'.
    If *dispw* is None, no clipping occurs.
    If *trunch* has a width greater than *dispw*, the empty string
    will be used as a truncator instead.'''
    if not s or (dispw is not None and dispw < 1): #iterator s would be truthy
        return '', 0

    w = 0
    ret = ''
    trunc_i = 0
    w_truncated = 0

    trunchlen = dispwidth(trunch)
    if dispw is None:
        s = ''.join(s)
        return s, dispwidth(s)
    if trunchlen > dispw: #if the truncator cannot fit, use a truncator of ''
        return _clipstr(s, dispw, trunch='', oddspacech=oddspacech, combch=combch, modch=modch)
    for c in s:
        newc, chlen = _dispch(c, oddspacech=oddspacech, combch=combch, modch=modch)
        if not newc:
            newc = c
            chlen = dispwidth(c)

        #if the next character will fit
        if w+chlen <= dispw:
            ret += newc
            w += chlen
            #move the truncation spot forward only when the truncation character can fit
            if w+trunchlen <= dispw:
                trunc_i += 1
                w_truncated += chlen
            continue
        # if we reach this line, a character did not fit, and the result needs truncation
        return ret[:trunc_i] + trunch, w_truncated+trunchlen

    return ret, w


@drawcache
def clipstr(s, dispw, truncator=None, oddspace=None):
    ''' *s* is a string or an iterator that contains characters.
    *dispw* is the integer screen width that the clipped string will fit inside, or None.'''
    if options.visibility:
        return _clipstr(s, dispw,
                        trunch=options.disp_truncator if truncator is None else truncator,
                        oddspacech=options.disp_oddspace if oddspace is None else oddspace,
                        modch='\u25e6',
                        combch='\u25cc')
    else:
        return _clipstr(s, dispw,
                trunch=options.disp_truncator if truncator is None else truncator,
                oddspacech=options.disp_oddspace if oddspace is None else oddspace,
                modch='',
                combch='')


def clipdraw(scr, y, x, s, attr, w=None, clear=True, literal=False, **kwargs):
    '''Draw `s`  at (y,x)-(y,x+w) with curses `attr`, clipping with ellipsis char.
       If `clear`, clear whole editing area before displaying.
       If `literal`, do not interpret internal color code markup.
       Return width drawn (max of w).
    '''
    if not literal:
        chunks = iterchunks(s, literal=literal)
    else:
        chunks = [(dict(link='', cattr=ColorAttr()), s)]

    x = max(0, x)
    y = max(0, y)

    return clipdraw_chunks(scr, y, x, chunks, attr, w=w, clear=clear, **kwargs)


def clipdraw_chunks(scr, y, x, chunks, cattr:ColorAttr=ColorAttr(), w=None, clear=True, literal=False, **kwargs):
    '''Draw `chunks` (sequence of (color:str, text:str) as from iterchunks) at (y,x)-(y,x+w) with curses `attr`, clipping with ellipsis char.
       If `clear`, clear whole editing area before displaying.
       Return width drawn (max of w).
    '''
    if scr:
        windowHeight, windowWidth = scr.getmaxyx()
    else:
        windowHeight, windowWidth = 25, 80
    totaldispw = 0

    assert isinstance(cattr, ColorAttr), cattr
    origattr = cattr
    origw = w
    clipped = ''
    link = ''

    if w and clear:
        actualw = min(w, windowWidth-x-1)
        if scr:
            scr.addstr(y, x, disp_column_fill*actualw, cattr.attr)  # clear whole area before displaying

    try:
        for colorstate, chunk in chunks:
            if colorstate:
                if isinstance(colorstate, str):
                    cattr = cattr.update(colors.get_color(colorstate), 100)
                else:
                    cattr = origattr.update(colorstate['cattr'], 100)
                    link = colorstate['link']

            if not chunk:
                continue

            if origw is None:
                chunkw = dispwidth(chunk, maxwidth=windowWidth-totaldispw)
            else:
                chunkw = origw-totaldispw

            chunkw = min(chunkw, windowWidth-x-1)
            if chunkw <= 0:  # no room anyway
                return totaldispw
            if not scr:
                return totaldispw

            # convert to string just before drawing
            clipped, dispw = clipstr(chunk, chunkw, **kwargs)

            if y >= 0 and y < windowHeight:
                scr.addstr(y, x, clipped, cattr.attr)
            else:
                if vd.options.debug:
                    raise Exception(f'addstr(y={y} x={x}) out of bounds')

            if link:
                vd.onMouse(scr, x, y, dispw, 1, BUTTON1_RELEASED=link)

            x += dispw
            totaldispw += dispw

            if chunkw < dispw:
                break
    except Exception as e:
        if vd.options.debug:
            raise
#        raise type(e)('%s [clip_draw y=%s x=%s dispw=%s w=%s clippedlen=%s]' % (e, y, x, totaldispw, w, len(clipped))
#                ).with_traceback(sys.exc_info()[2])

    return totaldispw


def _markdown_to_internal(text):
    'Return markdown-formatted `text` converted to internal formatting (like `[:color]text[/]`).'
    text = re.sub(r'`(.*?)`', r'[:code]\1[/]', text)
    text = re.sub(r'(^#.*?)$', r'[:heading]\1[/]', text)
    text = re.sub(r'\*\*(.*?)\*\*', r'[:bold]\1[/]', text)
    text = re.sub(r'\*(.*?)\*', r'[:italic]\1[/]', text)
    text = re.sub(r'\b_(.*?)_\b', r'[:underline]\1[/]', text)
    return text


def wraptext(text, width=80, indent=''):
    '''
    Word-wrap `text` and yield (formatted_line, textonly_line) for each line of at most `width` characters.
    Formatting like `[:color]text[/]` is ignored for purposes of computing width, and not included in `textonly_line`.
    '''
    import re

    if width <= 0:
        return

    for line in text.splitlines():
        if not line:
            yield '', ''
            continue

        line = _markdown_to_internal(line)
        chunks = re.split(internal_markup_re, line)
        textchunks = [x for x in chunks if not is_vdcode(x)]
        if ''.join(textchunks) == '':  #for markup with no contents, like '[:tag][/]' or '[:]' or '[/]'
            yield '', ''
            continue
        # textwrap.wrap does not handle variable-width characters  #2416
        for linenum, textline in enumerate(textwrap.wrap(''.join(textchunks), width=width, drop_whitespace=False)):
            txt = textline
            r = ''
            while chunks:
                c = chunks[0]
                if len(c) > len(txt):
                    r += txt
                    chunks[0] = c[len(txt):]
                    break

                if len(chunks) == 1:
                    r += chunks.pop(0)
                else:
                    chunks.pop(0)
                    r += txt[:len(c)] + chunks.pop(0)

                txt = txt[len(c):]

            r = r.strip()
            if linenum > 0:
                r = indent + r
            yield r, textline

        for c in chunks:
            yield c, ''


def clipbox(scr, lines, attr, title=''):
    scr.erase()
    scr.bkgd(attr)
    scr.box()
    h, w = scr.getmaxyx()
    for i, line in enumerate(lines):
        clipdraw(scr, i+1, 2, line, attr)

    clipdraw(scr, 0, w-dispwidth(title)-6, f"| {title} |", attr)

def clipstr_start(dispval, w, truncator=''):
    '''Return a tuple (frag, dw), where *frag* is the longest ending substring
    of *dispval* that will fit in a space *w* terminal display characters wide,
    and *dw* is the substring's display width as an int.'''
    # Note: this implementation is likely incorrect for unusual Unicode
    # strings or encodings, where trimming an initial character produces
    # an invalid string or does not make the string shorter.
    if w <= 0: return '', 0
    j = len(dispval)
    while j >= 1:
        if dispwidth((truncator if j > 1 else '') + dispval[j-1:]) <= w:
            j -= 1
        else:
            break
    frag = (truncator if j > 0 else '') + dispval[j:]
    return frag, dispwidth(frag)

def clipstr_middle(s, n=10, truncator='…'):
    '''Return a string having a display width <= *n*. Excess characters are
    trimmed from the middle of the string, and replaced by a single
    instance of *truncator*.'''
    if n == 0: return '', 0
    if dispwidth(s) > n:
        #for even widths, give the leftover 1 space to the right fragment
        l_space = n//2 if n%2 == 1 else max(n//2-1, 0)
        l_frag, l_w = _clipstr(s, l_space)
        #if left fragment did not fill its space, give the unused space to the right fragment
        r_frag = clipstr_start(s, n//2+(l_space-l_w))[0]
        res = l_frag + truncator + r_frag
        return res, dispwidth(res)
    return s, dispwidth(s)

vd.addGlobals(clipstr=clipstr,
              clipdraw=clipdraw,
              clipdraw_chunks=clipdraw_chunks,
              clipbox=clipbox,
              dispwidth=dispwidth,
              iterchars=iterchars,
              iterchunks=iterchunks,
              wraptext=wraptext,
              clipstr_start=clipstr_start,
              clipstr_middle=clipstr_middle)



import inspect

import visidata
from visidata import VisiData

@VisiData.api
def getStatusSource(vd):
    if not vd.options.debug:
        return ''
    stack = inspect.stack(context=0)  #2370
    for i, sf in enumerate(stack):
        if sf.function in 'status aside'.split():
            if stack[i+1].function in 'error fail warning debug'.split():
                sf = stack[i+2]
            else:
                sf = stack[i+1]
            break

    fn = sf.filename
    if fn.startswith(visidata.__path__[0]):
        fn = visidata.__package__ + fn[len(visidata.__path__[0]):]
    return f'{fn}:{sf.lineno}:{sf.function}'



from visidata import Column, vd, ColumnExpr, CompleteExpr, EscapeException, Sheet


@Column.api
def updateExpr(col, val):
    col.name = val
    try:
        col.expr = val
    except SyntaxError:
        col.expr = None

    col.sheet.draw(col.sheet._scr)


@Column.api  # expr.setter
def expr(self, expr):
    try:
        self.compiledExpr = compile(expr, '<expr>', 'eval') if expr else None
        self._expr = expr
    except SyntaxError as e:
        self._expr = None


@Sheet.api
def addcol_expr(sheet):
    try:
        c = sheet.addColumnAtCursor(ColumnExpr("", width=sheet.options.default_width))
        oldidx = sheet.cursorVisibleColIndex
        sheet.cursorVisibleColIndex = sheet.visibleCols.index(c)

        expr = sheet.editCell(sheet.cursorVisibleColIndex, -1,
                                completer=CompleteExpr(sheet),
                                updater=lambda val,col=c: col.updateExpr(val))

        c.expr = expr or vd.fail("no expr")
        c.name = expr
        c.width = None
    except (Exception, EscapeException):
        sheet.columns.remove(c)
        sheet.cursorVisibleColIndex = oldidx
        raise


Sheet.addCommand(None, 'addcol-expr', 'sheet.addcol_expr()', "create new column from Python expression, updating the column's calculated values live")
Sheet.addCommand(None, 'addcol-new', 'c=addColumnAtCursor(SettableColumn(name="", width=options.default_width)); draw(sheet._scr); cursorVisibleColIndex=visibleCols.index(c); c.name=editCell(cursorVisibleColIndex, -1); c.width=None', 'append new column, updating the column name live')



from visidata import *
from functools import wraps, partial


@asyncthread
@Sheet.api
def calcRowIndex(sheet, indexes):
    for rownum, r in enumerate(sheet.rows):
        indexes[sheet.rowid(r)] = rownum


@Sheet.lazy_property
def _rowindex(sheet):
    ret = {}
    sheet.calcRowIndex(ret)
    return ret


@Sheet.api
def rowindex(sheet, row):
    'Returns the rowindex given the row.  May spawn a thread to compute underlying _rowindex.'
    return sheet._rowindex.get(sheet.rowid(row))


@Sheet.api
def prev(sheet, row):
    'Return the row previous to the given row.'
    rownum = max(sheet.rowindex(row)-1, 0)
    return LazyComputeRow(sheet, sheet.rows[rownum])


@Sheet.api
def addcol_rowindex(sheet, newcol):
    oldAddRow = sheet.addRow
    def rownum_addRow(sheet, col, row, index=None):
        if index is None:
            index = len(sheet.rows)

        col._rowindex[sheet.rowid(row)] = index
        return oldAddRow(row, index)

    # wrapper addRow to keep the index up to date
    sheet.addRow = wraps(oldAddRow)(partial(rownum_addRow, sheet, newcol))
    sheet.addColumnAtCursor(newcol)

    # spawn a little thread to calc the rowindex
    sheet.calcRowIndex(newcol._rowindex)


@Sheet.api
def addcol_delta(sheet, vcolidx):
    col = sheet.visibleCols[vcolidx]

    newcol = ColumnExpr("delta_"+col.name,
                   type=col.type,
                   _rowindex={},    # [rowid(row)] -> rowidx
                   expr="{0}-prev(row).{0}".format(col.name))

    sheet.addcol_rowindex(newcol)
    return newcol

@Sheet.api
def addcol_rownum(sheet):
    newcol = Column("rownum",
               type=int,
               _rowindex={},    # [rowid(row)] -> rowidx
               getter=lambda col,row: col._rowindex.get(col.sheet.rowid(row)))

    sheet.addcol_rowindex(newcol)
    return newcol

Sheet.addCommand(None, 'addcol-rownum', 'addcol_rownum()', helpstr='add column with original row ordering')
Sheet.addCommand(None, 'addcol-delta', 'addcol_delta(cursorVisibleColIndex)', helpstr='add column with delta of current column')



from visidata import vd, Column, Sheet, asyncthread, Progress


class HistogramColumn(Column):
    def calcValue(col, row):
        histogram = col.sheet.options.disp_histogram
        histolen = (col.width-2)*col.sourceCol.getTypedValue(row)//(col.sourceCol.largest-col.sourceCol.smallest)
        return histogram*histolen


@Sheet.api
def addcol_histogram(sheet, col):  #2052
    newcol = HistogramColumn(col.name+'_histogram', sourceCol=col)
    col.smallest = None
    col.largest = None
    sheet.calc_histogram_bounds(col)
    return newcol


@Sheet.api
@asyncthread
def calc_histogram_bounds(sheet, col):
    for row in Progress(sheet.rows):
        v = col.getTypedValue(row)
        if col.smallest is None or v < col.smallest:
            col.smallest = v
        if col.largest is None or v > col.largest:
            col.largest = v


Sheet.addCommand('', 'addcol-histogram', 'addColumnAtCursor(addcol_histogram(cursorCol))', 'add column with histogram of current column')


vd.addMenuItems('Column > Add column > histogram > addcol-histogram')



import json

from visidata import VisiData, vd, SequenceSheet, deduceType, Progress


@VisiData.api
def guess_jsonla(vd, p):
    '''A JSONLA file is a JSONL file with rows of arrays, where the first row
    is a header array:

    ["A", "B", "C"]
    [1, "blue", true]
    [2, "yellow", false]

    The header array must be a flat array of strings

    If no suitable header is found, fall back to generic JSON load.
    '''

    with p.open(encoding=vd.options.encoding) as fp:
        try:
            first_line = next(fp)
        except StopIteration:
            return

    if first_line.strip().startswith('['):
        try:
            ret = json.loads(first_line)
        except json.decoder.JSONDecodeError:
            return
        if isinstance(ret, list) and all(isinstance(v, str) for v in ret):
            return dict(filetype='jsonla')


@VisiData.api
def open_jsonla(vd, p):
    return JsonlArraySheet(p.base_stem, source=p)


class JsonlArraySheet(SequenceSheet):
    rowtype = 'rows'    # rowdef: list of Python objects decoded from JSON
    def iterload(self):
        with self.open_text_source() as fp:
            for L in fp:
                yield json.loads(L)

        # set column types from first row
        for i, c in enumerate(self.columns):
            c.type = deduceType(self.rows[0][i])


def get_jsonla_rows(sheet, cols):
    for row in Progress(sheet.rows):
        yield [vd.get_json_value(col, row) for col in cols]


class _vjsonEncoder(json.JSONEncoder):
    def default(self, obj):
        return str(obj)


def write_jsonla(vs, fp):
        vcols = vs.visibleCols
        jsonenc = _vjsonEncoder()
        with Progress(gerund='saving'):
            header = [col.name for col in vcols]
            fp.write(jsonenc.encode(header) + '\n')
            rows = get_jsonla_rows(vs, vcols)
            for row in rows:
                fp.write(jsonenc.encode(row) + '\n')


@VisiData.api
def save_jsonla(vd, p, *vsheets):
    with p.open(mode='w', encoding=vsheets[0].options.save_encoding) as fp:
        for vs in vsheets:
            write_jsonla(vs, fp)


JsonlArraySheet.options.regex_skip = r'^(//|#).*'



from visidata import vd, options, TypedWrapper, asyncthread, Progress
from visidata import wrapply, VisiData

vd.option('graphviz_edge_labels', True, 'whether to include edge labels on graphviz diagrams')


def is_valid(v):
    if v is None:
        return False
    if isinstance(v, TypedWrapper):
        return False
    return True


@VisiData.api
def save_dot(vd, p, vs):
    unusedColors = 'orange green purple cyan red blue black'.split()
    assignedColors = {}

    srccol = vs.keyCols[0]
    dstcol = vs.keyCols[1]
    with p.open(mode='w', encoding='utf-8') as fp:
        pfp = lambda *args: print(*args, file=fp)
        pfp('graph { concentrate=true;')
        for row in Progress(vs.rows, 'saving'):
            src = srccol.getTypedValue(row)
            dst = dstcol.getTypedValue(row)
            if not is_valid(src) or not is_valid(dst):
                continue

            downsrc = vd.cleanName(str(src)) or src
            downdst = vd.cleanName(str(dst)) or dst
            edgenotes = [c.getTypedValue(row) for c in vs.nonKeyVisibleCols if not vd.isNumeric(c)]
            edgetype = '-'.join(str(x) for x in edgenotes if is_valid(x))
            color = assignedColors.get(edgetype, None)
            if not color:
                color = unusedColors.pop() if unusedColors else 'black'
                assignedColors[edgetype] = color

            if options.graphviz_edge_labels:
                nodelabels = [wrapply(vd.SIFormatter, '%0.1f', c.getTypedValue(row)) for c in vs.nonKeyVisibleCols if vd.isNumeric(c)]
                label = '/'.join(str(x) for x in nodelabels if is_valid(x))
            else:
                label = ''
            pfp('\t%s[label="%s"];' % (downsrc, src))
            pfp('\t%s[label="%s"];' % (downdst, dst))
            pfp('\t%s -- %s[label="%s", color=%s];' % (downsrc, downdst, label, color))

        pfp('label="%s"' % vs.name)
        pfp('node[shape=plaintext];')
        pfp('subgraph cluster_legend {')
        pfp('label="Legend";')
        for i, (k, color) in enumerate(assignedColors.items()):
            pfp('key%d[label="%s", fontcolor=%s];' % (i, k, color))

        pfp('}')  # legend subgraph
        pfp('}')



#!/usr/bin/env python3

__all__=[ 'SelectorColumn', 'soupstr' ]

import os.path
from urllib.parse import urljoin

import concurrent.futures
import functools

from visidata import vd, VisiData, TableSheet, vdtype, Column, AttrColumn, Progress, date


@VisiData.api
def soup(vd, s):
    bs4 = vd.importExternal('bs4', 'beautifulsoup4')
    from bs4 import BeautifulSoup
    return BeautifulSoup(s, 'html.parser')


@VisiData.api
def open_scrape(vd, p):
    bs4 = vd.importExternal('bs4', 'beautifulsoup4')

    vd.enable_requests_cache()
    if p.is_url():
        return HtmlDocsSheet(p.base_stem, source=p, urls=[p.given])
    else:
        return HtmlElementsSheet(p.base_stem, source=p, elements=None)

VisiData.openhttp_scrape = VisiData.open_scrape

def node_name(node):
    me = node.name
    class_ = node.attrs.get("class")
    if class_:
        me += '.' + class_[0]
    id_ = node.attrs.get("id")
    if id_:
        me += '#' + id_
    return me

@functools.lru_cache(maxsize=None)
def calc_selector(node):
    if not node.parent:
        return ''

    psel = calc_selector(node.parent)
    oursel = node_name(node)
    if not psel:
        return oursel

    root = list(node.parents)[-1]

    combinedsel = psel+' '+oursel
    if len(root.select(combinedsel)) == len(root.select(oursel)):
        return oursel

    return combinedsel


class HtmlAttrColumn(Column):
    def calcValue(self, row):
        return row.attrs.get(self.expr)


def prev_header(r):
    hdrtags = ['h1', 'h2', 'h3', 'h4', 'h5', 'h6']
    try:
        i = hdrtags.index(r.name)
        return r.find_previous(hdrtags[:i-1])
    except Exception:
        return r.find_previous(hdrtags)


# one row per element
class HtmlElementsSheet(TableSheet):
    guide = '''# HTMLElements

This is a list of HTML elements from _{sheet.source}_ as parsed by `beautifulsoup4`.

Standard VisiData exploration techniques can be used to find relevant data, which will help determine the proper selector to use.

- `Enter` to dive into children of cursor element (or children of all selected rows with `g Enter`)
- `go` to batch open links in selected rows on new RequestsSheet, which will fetch each page
- `~` to use the `soupstr` type to join all the text elements
'''
    # source=[element, ...]
    rowtype='dom nodes'  # rowdef soup.element
    columns = [
        Column('name', getter=lambda c,r: node_name(r)),
        Column('selector', getter=lambda c,r: calc_selector(r), cache='async', width=0),
        AttrColumn('string'),
        Column('depth', cache=True, getter=lambda c,r: list(c.sheet.html_parents(r))),
        Column('prev_header', getter=lambda c,r: prev_header(r), cache=True),
        HtmlAttrColumn('href', expr='href'),
    ]
    def iterload(self):
        for el in self.elements or [vd.soup(self.source.read_text())]:
            for x in el.find_all():
                if x.string:
                    yield x

    def html_parents(self, row):
        while row.parent and row.parent is not row:
            yield row.parent
            row = row.parent

    @property
    def rootSource(self):
        return self.rootSheet.source

    def openRows(self, rows):
        realurls = [urljoin(self.rootSource.given, r.attrs.get('href')) for r in rows]
        yield HtmlDocsSheet(self.name, 'scrape', source=self, urls=realurls)

    def openRow(self, row):
        'opening a single row'
        return HtmlElementsSheet('', source=self, elements=[row])


class DocsSelectorColumn(Column):
    def calcValue(self, row):
        return [x for x in row.soup.select(self.expr)]

class SelectorColumn(Column):
    def calcValue(self, row):
        return [x for x in row.select(self.expr)]


# urls=list of urls to scrape
class HtmlDocsSheet(TableSheet):
    help='''# HtmlDocsSheet

- `Enter` to open the current request as list of HTMLElements
- `;` to add column of elements matching given css selector
  - this is how to cross-tabulate data from multiple pages
'''
    rowtype='requests'  # rowdef: requests.Response
    columns = [
        AttrColumn('url'),
        AttrColumn('status_code', type=int),
        AttrColumn('from_cache'),
        AttrColumn('fetched_at', 'created_at', type=date, width=0),
        AttrColumn('expires', type=date),
        AttrColumn('reason'),
        AttrColumn('soup.title.string'),
    ]
    def iterload(self):
        requests = vd.importExternal('requests')
        self.colnames = {}
#        with concurrent.futures.ThreadPoolExecutor(max_workers=8) as executor:
#            yield from executor.map(requests.get, Progress(self.urls))
        for url in Progress(self.urls):
            yield requests.get(url)

    def addRow(self, row, index=None):
        super().addRow(row, index=index)
        row.soup = vd.callNoExceptions(vd.soup, row.text)

    def openRow(self, row):
        return HtmlElementsSheet(row.url, source=self, elements=[row.soup])

def soupstr(coll):
    return ' '.join(v.string for v in coll)

vdtype(soupstr, 's')

@TableSheet.api
def scrape_urls(sheet, col, rows):
    return HtmlDocsSheet(sheet.name, "selected_urls", urls=[col.getTypedValue(r) for r in rows])

HtmlElementsSheet.addCommand('~', 'type-soupstr', 'cursorCol.type=soupstr', 'set type of current column to list of html elements')
HtmlElementsSheet.addCommand('go', 'open-rows', 'for vs in openRows(selectedRows): vd.push(vs)', 'open sheet for each selected element')
TableSheet.addCommand('gzo', 'scrape-cells', 'vd.push(scrape_urls(cursorCol, selectedRows))', 'open HTML Documents sheet from selected URLs')
HtmlDocsSheet.addCommand(';', 'addcol-selector', 'sel=input("css selector: ", type="selector"); addColumn(DocsSelectorColumn(sel, expr=sel, cache="async"))', 'add column derived from css selector of current column')
HtmlElementsSheet.addCommand(';', 'addcol-selector', 'sel=input("css selector: ", type="selector"); addColumn(SelectorColumn(sel, expr=sel, cache="async"))',  'add column derived from css selector of current column')

vd.addGlobals({
    'HtmlDocsSheet':SelectorColumn,
    'SelectorColumn':SelectorColumn,
    'DocsSelectorColumn':DocsSelectorColumn,
})

vd.addMenuItem('Data', '+Scrape', 'selected cells', 'scrape-cells')



import re

from visidata import vd, BaseSheet

vd.option('regex_skip', '', 'regex of lines to skip in text sources', help='regex', replay=True)
vd.option('regex_flags', 'I', 'flags to pass to re.compile() [AILMSUX]', replay=True)

@BaseSheet.api
def regex_flags(sheet):
    'Return flags to pass to regex functions from options'
    return sum(getattr(re, f.upper()) for f in sheet.options.regex_flags)


class FilterFile:
    def __init__(self, fp, regex:str, regex_flags:int=0):
        import re
        self._fp = fp
        self._regex_skip = re.compile(regex, regex_flags)

    def readline(self) -> str:
        while True:
            line = self._fp.readline()
            if self._regex_skip.match(line):
                continue
            return line

    def __getattr__(self, k):
        return getattr(self._fp, k)

    def __iter__(self):
        return self

    def __next__(self):
        line = self.readline()
        if not line:
            raise StopIteration
        return line

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        return self._fp.__exit__(*args, **kwargs)


@BaseSheet.api
def open_text_source(sheet, **kwargs):
    'Open sheet source as text, passing **kwargs to .open() (default to sheet options for encoding and regex_skip).'
    openkwargs = dict(encoding=sheet.options.encoding,
                      encoding_errors=sheet.options.encoding_errors)
    openkwargs.update(kwargs)
    fp = sheet.source.open(**openkwargs)
    regex_skip = sheet.options.regex_skip
    if regex_skip:
        return FilterFile(fp, regex_skip, sheet.regex_flags())
    return fp



#!/usr/bin/python3

from visidata import vd, VisiData, JsonSheet, ColumnAttr, Path, ENTER, AttrDict, ExpectedException, stacktrace, TypedExceptionWrapper
import json
import os
from os import linesep

@VisiData.api
def open_grep(vd, p):
    return GrepSheet(p.base_stem, source=p)

@VisiData.api
def save_grep(vd, p, *vsheets):
    vd.save_jsonl(p, *vsheets)

def format_row(rowdict):
    # handle rows that are output of 'rg --json'
    if 'type' in rowdict and rowdict['type'] == 'match':
        match_data = rowdict['data']
        d = {
            'file':    match_data['path']['text'],
            'line_no': match_data['line_number'],
            'text':    match_data['lines']['text'].rstrip(linesep)
        }
        return AttrDict(d)
    # handle a .grep file that was saved by visidata, or
    # ripgrep rows that were preprocessed by jq:  'rg --json |jq [...]'
    if 'line_no' in rowdict:
        rowdict['text'] = rowdict['text'].rstrip(linesep)
        return AttrDict(rowdict)
    return None

class GrepSheet(JsonSheet):
    # The input file is in JSON Lines format, where each line describes a JSON object.
    # The JSON objects are either in the ripgrep grep_printer format:
    # https://docs.rs/grep-printer/0.1.0/grep_printer/struct.JSON.html
    # or contain the keys 'file', 'line_no', and 'text'.
    _rowtype = 'lines'  # rowdef: AttrDict

    columns = [
        ColumnAttr('file', type=str),
        ColumnAttr('line_no', type=int),
        ColumnAttr('text', type=str)
    ]
    nKeys = 2
    def iterload(self):
        with self.open_text_source() as fp:
            for L in fp:
                try:
                    if not L: # skip blank lines
                        continue
                    json_obj = json.loads(L)
                    if not isinstance(json_obj, dict):
                        vd.fail(f'line does not hold a JSON object:  {L}')
                    row = format_row(json_obj)
                    if not row: #skip lines that do not contain match data
                        continue
                    yield row
                except ValueError as e:
                    if self.rows:   # if any rows have been added already
                        e.stacktrace = stacktrace()
                        yield TypedExceptionWrapper(json.loads, L, exception=e)  # an error on one line
                    else:
                        # If input is not JSON, parse it as output of 'grep -n':  file:line_no:text
                        # If that does not parse, parse it as output of typical 'grep':  file:text
                        with self.open_text_source() as fp:
                            try:
                                extract_line_no = True
                                for L in fp:
                                    L = L.rstrip(linesep)
                                    sep1 = L.index(':')
                                    if extract_line_no:
                                        sep2 = L.find(':', sep1+1)
                                        try:
                                            if sep2 == -1: raise ValueError
                                            line_no = int(L[sep1+1:sep2]) # may raise ValueError
                                            if line_no < 1: raise ValueError
                                            text = L[sep2+1:]
                                        except ValueError: # if we can't find a line_no that is > 0, with a separator after it
                                            extract_line_no = False
                                            line_no = None
                                            text = L[sep1+1:]
                                    else:
                                        text = L[sep1+1:]
                                    yield AttrDict({'file':    L[:sep1],
                                                    'line_no': line_no,
                                                    'text':    text})
                            except ValueError:
                                vd.fail('file is not grep output')
                        break

    def afterLoad(self):
        if self.nRows == 0:
            vd.status('no grep results found in input data')

@GrepSheet.api
def sysopen_row(sheet, row):
    '''Open the file in an editor at the specific line.'''
    if sheet.nRows == 0: return
    try:
        given = row.file
        if vd.options.grep_base_dir and not os.path.isabs(given):
            given = vd.options.grep_base_dir + os.sep + row.file
        p = Path(given)
    except TypeError:
        vd.fail(f'cannot open row: {given}')
    if p.exists():
        # works for vim and emacsclient
        if row.line_no is not None:
            vd.launchEditor(p.given, f'+{row.line_no:d}')
        else:
            vd.launchEditor(p.given)
    else:
        vd.fail(f'cannot find file: {p.given}')

GrepSheet.addCommand(ENTER, 'sysopen-row', 'sysopen_row(cursorRow)', 'open current file in external $EDITOR, at the line')

vd.addGlobals({
    'GrepSheet': GrepSheet,
})
vd.option('grep_base_dir', None, 'base directory for relative paths opened with sysopen-row')



import time

from visidata import Progress, Sheet, Column, asyncthread, vd, Column


class ExprColumn(Column):
    'Column using *expr* to derive the value from each row.'
    def __init__(self, name, expr=None, **kwargs):
        super().__init__(name, **kwargs)
        self.expr = expr or name
        self.ncalcs = 0
        self.totaltime = 0
        self.maxtime = 0

    def calcValue(self, row):
        t0 = time.perf_counter()
        r = self.sheet.evalExpr(self.compiledExpr, row, col=self, curcol=self)
        t1 = time.perf_counter()
        self.ncalcs += 1
        self.maxtime = max(self.maxtime, t1-t0)
        self.totaltime += (t1-t0)
        return r

    def putValue(self, row, val):
        a = self.getDisplayValue(row)
        b = self.format(self.type(val))
        if a != b:
            vd.warning("Cannot change value of calculated column.  Use `'` to freeze column.")

    @property
    def expr(self):
        return self._expr

    @expr.setter
    def expr(self, expr):
        self.compiledExpr = compile(expr, '<expr>', 'eval') if expr else None
        self._expr = expr


class CompleteExpr:
    def __init__(self, sheet=None):
        self.varnames = []
        if sheet:
            self.varnames.extend(sorted(col.name for col in sheet.columns))
        self.varnames.extend(sorted(x for x in vd.getGlobals()))
        for c in vd.contexts:
            self.varnames.extend(sorted(x for x in dir(c)))

    def __call__(self, val, state):
        i = len(val)-1
        while val[i:].isidentifier() and i >= 0:
            i -= 1

        if i < 0:
            base = ''
            partial = val
        elif val[i] == '.':  # no completion of attributes
            return None
        else:
            base = val[:i+1]
            partial = val[i+1:]

        # Remove unmatching and duplicate completions
        varnames_dict = {x:None for x in self.varnames if x.startswith(partial)}
        varnames = list(varnames_dict.keys())

        if not varnames:
            return val

        return base + varnames[state%len(varnames)]


@Column.api
@asyncthread
def setValuesFromExpr(self, rows, expr, **kwargs):
    'Set values in this column for *rows* to the result of the Python expression *expr* applied to each row.'
    compiledExpr = compile(expr, '<expr>', 'eval')
    vd.addUndoSetValues([self], rows)
    nset = 0
    for row in Progress(rows, 'setting'):
        # Note: expressions that are only calculated once, do not need to pass column identity
        # they can reference their "previous selves" once without causing a recursive problem
        try:
            v = self.sheet.evalExpr(compiledExpr, row, **kwargs)
            self.setValue(row, v)
            nset += 1
        except Exception as e:
            vd.exceptionCaught(e)

    self.recalc()
    vd.status(f'set {nset} values = {expr}')


@Sheet.api
def inputExpr(self, prompt, *args, **kwargs):
    return vd.input(prompt, "expr", *args, completer=CompleteExpr(self), **kwargs)


Sheet.addCommand('=', 'addcol-expr', 'addColumnAtCursor(ExprColumn(inputExpr("new column expr="), col=cursorCol, curcol=cursorCol))', 'create new column from Python expression, with column names as variables')
Sheet.addCommand('g=', 'setcol-expr', 'cursorCol.setValuesFromExpr(someSelectedRows, inputExpr("set selected="), curcol=cursorCol)', 'set current column for selected rows to result of Python expression')
Sheet.addCommand('z=', 'setcell-expr', 'cursorCol.setValues([cursorRow], evalExpr(inputExpr("set expr="), row=cursorRow, curcol=cursorCol))', 'evaluate Python expression on current row and set current cell with result of Python expression')
Sheet.addCommand('gz=', 'setcol-iter', 'cursorCol.setValues(someSelectedRows, *list(itertools.islice(eval(input("set column= ", "expr", completer=CompleteExpr())), len(someSelectedRows))))', 'set current column for selected rows to the items in result of Python sequence expression')
Sheet.addCommand('', 'addcol-iter', 'iter_expr=inputExpr("new column iterator expr: "); it = eval(iter_expr, getGlobals()); c=SettableColumn(); addColumnAtCursor(c); c.setValues(rows, *it)', 'add column with values from a Python sequence expression, repeating it if needed to fill')

Sheet.addCommand(None, 'show-expr', 'status(evalExpr(inputExpr("show expr="), row=cursorRow, curcol=cursorCol))', 'evaluate Python expression on current row and show result on status line')

vd.addGlobals(
    ExprColumn=ExprColumn,
    ColumnExpr=ExprColumn,
    CompleteExpr=CompleteExpr,
)

vd.addMenuItems('''
    Edit > Modify > current cell > Python expression > setcell-expr
    Edit > Modify > selected cells > Python sequence > setcol-expr
    Column > Add column > Python expr > addcol-expr
''')



# requirements: mutagen

import functools

from visidata import vd, Column, AttrColumn, DirSheet


@functools.lru_cache(None)
def get_mutagen_info(path):
    mutagen = vd.importExternal('mutagen')
    m = mutagen.File(path)
    return m.info


class MutagenColumn(AttrColumn):
    def calcValue(self, r):
        md = get_mutagen_info(r)
        return getattr(md, self.expr, None)


@DirSheet.api
def audiometadata_columns(sheet):
    return [
        Column('audio_info', width=0, getter=lambda c,r: get_mutagen_info(r)),
        MutagenColumn('bitrate'),
        MutagenColumn('channels'),
        MutagenColumn('encoder_info'),
        MutagenColumn('encoder_settings'),
        MutagenColumn('frame_offset'),
        MutagenColumn('length'),
        MutagenColumn('mode'),
        MutagenColumn('padding'),
        MutagenColumn('protected'),
        MutagenColumn('sample_rate'),
        MutagenColumn('track_gain'),
    ]


DirSheet.addCommand('', 'addcol-audiometadata', 'addColumn(*audiometadata_columns())', 'add metadata columns for audio files (MP3, FLAC, Ogg, etc)')


vd.addMenuItems('Column > Add column > audio metadata > addcol-audiometadata')



from visidata import VisiData, vd, Column, Sheet, Fanout, asyncthread

@Column.api
def setWidth(self, w):
    if self.width != w:
        if self.width == 0 or w == 0:  # hide/unhide
            vd.addUndo(setattr, self, '_width', self.width)
    self._width = w


@Column.api
def toggleWidth(self, width):
    'Change column width to either given `width` or default value.'
    if self.width != width:
        self.width = width
    else:
        self.width = int(self.sheet.options.default_width)


@Column.api
def toggleMultiline(self):
    if self.height == 1:
        self.height = self.sheet.options.default_height
    else:
        self.height = 1

@VisiData.api
def unhide_cols(vd, cols, rows):
    'sets appropriate width if column was either hidden (0) or unseen (None)'
    for c in cols:
        c.setWidth(abs(c.width or 0) or c.getMaxWidth(rows))

@VisiData.api
def hide_col(vd, col):
    if not col: vd.fail("no columns to hide")
    col.hide()

@Sheet.api
@asyncthread
def hide_uniform_cols(sheet):
    if len(sheet.rows) < 2:
        return
    idx = sheet.cursorVisibleColIndex
    for i, col in enumerate(sheet.visibleCols):
        vals = (col.getTypedValue(r) for r in sheet.rows)
        first = next(vals)
        if all(v == first for v in vals):
            vd.status(f'hide {col.name} with value: {col.getDisplayValue(sheet.rows[0])}')
            if i <= idx:
                sheet.cursorRight(-1)
            col.hide()

Sheet.addCommand('_', 'resize-col-max', 'if cursorCol: cursorCol.toggleWidth(cursorCol.getMaxWidth(visibleRows))', 'toggle width of current column between full and default width')
Sheet.addCommand('z_', 'resize-col-input', 'width = int(input("set width= ", value=cursorCol.width)); cursorCol.setWidth(width)', 'adjust width of current column to N')
Sheet.addCommand('g_', 'resize-cols-max', 'for c in visibleCols: c.toggleWidth(c.getMaxWidth(visibleRows))', 'toggle widths of all visible columns between full and default width')
Sheet.addCommand('gz_', 'resize-cols-input', 'width = int(input("set width= ", value=cursorCol.width)); Fanout(visibleCols).setWidth(width)', 'adjust widths of all visible columns to N')

Sheet.addCommand('-', 'hide-col', 'hide_col(cursorCol)', 'hide the current column')
Sheet.addCommand('z-', 'resize-col-half', 'width = cursorCol.width if cursorCol.width is not None else options.default_width; cursorCol.setWidth(width//2)', 'reduce width of current column by half')
Sheet.addCommand('g-', 'hide-uniform-cols', 'sheet.hide_uniform_cols()', 'hide any column that has multiple rows but only one distinct value')

Sheet.addCommand('gv', 'unhide-cols', 'unhide_cols(columns, visibleRows)', 'unhide all hidden columns on current sheet')
Sheet.addCommand('v', 'toggle-multiline', 'for c in visibleCols: c.toggleMultiline()', 'toggle multiline display')
Sheet.addCommand('zv', 'resize-height-input', 'Fanout(visibleCols).height=int(input("set height for all columns to: ", value=max(c.height for c in sheet.visibleCols)))', 'resize row height to N')
Sheet.addCommand('gzv', 'resize-height-max', 'h=calc_height(cursorRow, {}, maxheight=windowHeight-1); vd.status(f"set height for all columns to {h}"); Fanout(visibleCols).height=h', 'resize row height to max height needed to see this row')

vd.addMenuItems('''
    Column > Hide > Hide > hide-col
    Column > Hide > Hide uniform > hide-uniform-cols
    Column > Hide > Unhide all > unhide-cols
    Column > Resize > half width > resize-col-half
    Column > Resize > current column width to max > resize-col-max
    Column > Resize > current column width to N > resize-col-input
    Column > Resize > all columns width to max > resize-cols-max
    Column > Resize > all columns width to N > resize-cols-input
    Row > Resize > height to N > resize-height-input
    Row > Resize > height to max > resize-height-max
    View > Toggle display > multiline > toggle-multiline
''')



import threading

from visidata import vd, UNLOADED, namedlist, vlen, asyncthread, globalCommand, date
from visidata import VisiData, BaseSheet, Sheet, ColumnAttr, VisiDataMetaSheet, JsonLinesSheet, TypedWrapper, AttrDict, Progress, ErrorSheet, CompleteKey, Path
import visidata

vd.option('replay_wait', 0.0, 'time to wait between replayed commands, in seconds', sheettype=None)
vd.theme_option('disp_replay_play', '▶', 'status indicator for active replay')
vd.theme_option('disp_replay_record', '⏺', 'status indicator for macro record')
vd.theme_option('color_status_replay', 'green', 'color of replay status indicator')

# prefixes which should not be logged
nonLogged = '''forget exec-longname undo redo quit
show error errors statuses options threads jump
replay cancel save-cmdlog macro cmdlog-sheet menu repeat reload-every
search scroll prev next page start end zoom visibility sidebar
mouse suspend redraw no-op help syscopy sysopen profile toggle'''.split()

vd.option('rowkey_prefix', 'キ', 'string prefix for rowkey in the cmdlog', sheettype=None)

vd._nextCommands = []  # list[str|CommandLogRow] for vd.queueCommand

CommandLogRow = namedlist('CommandLogRow', 'sheet col row longname input keystrokes comment undofuncs'.split())

@VisiData.api
def queueCommand(vd, longname, input=None, sheet=None, col=None, row=None):
    'Add command to queue of next commands to execute.'
    vd._nextCommands.append(CommandLogRow(longname=longname, input=input, sheet=sheet, col=col, row=row))


@VisiData.api
def open_vd(vd, p):
    return CommandLog(p.base_stem, source=p, precious=True)

@VisiData.api
def open_vdj(vd, p):
    return CommandLogJsonl(p.base_stem, source=p, precious=True)

VisiData.save_vd = VisiData.save_tsv


@VisiData.api
def save_vdj(vd, p, *vsheets):
    with p.open(mode='w', encoding=vsheets[0].options.save_encoding) as fp:
        fp.write("#!vd -p\n")
        for vs in vsheets:
            vs.write_jsonl(fp)


@VisiData.api
def checkVersion(vd, desired_version):
    if desired_version != visidata.__version_info__:
        vd.fail("version %s required" % desired_version)

@VisiData.api
def fnSuffix(vd, prefix:str):
    i = 0
    fn = prefix + '.vdj'
    while Path(fn).exists():
        i += 1
        fn = f'{prefix}-{i}.vdj'

    return fn

def indexMatch(L, func):
    'returns the smallest i for which func(L[i]) is true'
    for i, x in enumerate(L):
        if func(x):
            return i

@VisiData.api
def isLoggableCommand(vd, cmd):
    'Return whether command should be logged to the cmdlog, depending if it has a prefix in nonLogged, or was defined with replay=False.'
    if not cmd.replayable:
        return False

    for n in nonLogged:
        if cmd.longname.startswith(n):
            return False
    return True

def isLoggableSheet(sheet):
    return sheet is not vd.cmdlog and not isinstance(sheet, (vd.OptionsSheet, ErrorSheet))


@Sheet.api
def moveToRow(vs, rowstr):
    'Move cursor to row given by *rowstr*, which can be either the row number or keystr.'
    rowidx = vs.getRowIndexFromStr(rowstr)
    if rowidx is None:
        return False

    vs.cursorRowIndex = rowidx

    return True

@Sheet.api
def getRowIndexFromStr(vs, row):
    prefix = vd.options.rowkey_prefix
    index = None
    if isinstance(row, int):
        index = row
    elif isinstance(row, str) and row.startswith(prefix):
        rowk = row[len(prefix):]
        index = indexMatch(vs.rows, lambda r,vs=vs,rowk=rowk: rowk == ','.join(map(str, vs.rowkey(r))))
    else:
        try:
            index = int(row)
        except ValueError:
            vd.warning('invalid type for row index')

    return index


@Sheet.api
def moveToCol(vs, col):
    'Move cursor to column given by *col*, which can be either the column number or column name.'
    if isinstance(col, str):
        vcolidx = indexMatch(vs.visibleCols, lambda c,name=col: name == c.name)
    elif isinstance(col, int):
        vcolidx = col

    if vcolidx is None or vcolidx >= vs.nVisibleCols:
        return False

    vs.cursorVisibleColIndex = vcolidx

    return True


@BaseSheet.api
def commandCursor(sheet, execstr):
    'Return (col, row) of cursor suitable for cmdlog replay of execstr.'
    colname, rowname = '', ''
    contains = lambda s, *substrs: any((a in s) for a in substrs)
    if contains(execstr, 'cursorTypedValue', 'cursorDisplay', 'cursorValue', 'cursorCell', 'cursorRow') and sheet.nRows > 0:
        rowname = sheet.cursorRowIndex

    if contains(execstr, 'cursorTypedValue', 'cursorDisplay', 'cursorValue', 'cursorCell', 'cursorCol', 'cursorVisibleCol', 'ColumnAtCursor'):
        if sheet.cursorCol:
            colname = sheet.cursorCol.name or sheet.visibleCols.index(sheet.cursorCol)
        else:
            colname = None
    return colname, rowname


# rowdef: namedlist (like TsvSheet)
class CommandLogBase:
    'Log of commands for current session.'
    rowtype = 'logged commands'
    precious = False
    _rowtype = CommandLogRow
    columns = [
        ColumnAttr('sheet'),
        ColumnAttr('col'),
        ColumnAttr('row'),
        ColumnAttr('longname'),
        ColumnAttr('input'),
        ColumnAttr('keystrokes'),
        ColumnAttr('comment'),
        ColumnAttr('undo', 'undofuncs', type=vlen, width=0)
    ]

    filetype = 'vd'

    def newRow(self, **fields):
        return self._rowtype(**fields)

    def beforeExecHook(self, sheet, cmd, args, keystrokes):
        if vd.activeCommand:
            self.afterExecSheet(sheet, False, '')

        colname, rowname, sheetname = '', '', None
        if sheet and not (cmd.longname.startswith('open-') and not cmd.longname in ('open-row', 'open-cell')):
            sheetname = sheet.name

            colname, rowname = sheet.commandCursor(cmd.execstr)

            contains = lambda s, *substrs: any((a in s) for a in substrs)
            if contains(cmd.execstr, 'pasteFromClipboard'):
                args = vd.sysclipValue().strip()


        comment = vd.currentReplayRow.comment if vd.currentReplayRow else cmd.helpstr
        vd.activeCommand = self.newRow(sheet=sheetname,
                                            col=colname,
                                            row=rowname,
                                            keystrokes=keystrokes,
                                            input=args,
                                            longname=cmd.longname,
                                            comment=comment,
                                            replayable=cmd.replayable,
                                            undofuncs=[])

    def afterExecSheet(self, sheet, escaped, err):
        'Records vd.activeCommand'
        if not vd.activeCommand:  # nothing to record
            return

        if escaped:
            vd.activeCommand = None
            return

        # remove user-aborted commands and simple movements (unless first command on the sheet, which created the sheet)
        if not sheet.cmdlog_sheet.rows or vd.isLoggableCommand(vd.activeCommand):
            if isLoggableSheet(sheet):      # don't record actions from cmdlog or other internal sheets on global cmdlog
                self.addRow(vd.activeCommand)  # add to global cmdlog
            sheet.cmdlog_sheet.addRow(vd.activeCommand)  # add to sheet-specific cmdlog

        vd.activeCommand = None

    def openHook(self, vs, src):
        while isinstance(src, BaseSheet):
            src = src.source
        r = self.newRow(keystrokes='o', input=str(src), longname='open-file', replayable=True)
        vs.cmdlog_sheet.addRow(r)
        self.addRow(r)

class CommandLog(CommandLogBase, VisiDataMetaSheet):
    pass

class CommandLogJsonl(CommandLogBase, JsonLinesSheet):

    filetype = 'vdj'

    def newRow(self, **fields):
        return AttrDict(JsonLinesSheet.newRow(self, **fields))

    def iterload(self):
        for r in JsonLinesSheet.iterload(self):
            if isinstance(r, TypedWrapper):
                yield r
            else:
                yield AttrDict(r)


### replay

vd.paused = False
vd.currentReplay = None     # CommandLog replaying currently
vd.currentReplayRow = None  # must be global, to allow replay


@VisiData.api
def replay_cancel(vd):
    vd.currentReplayRow = None
    vd.currentReplay = None
    vd._nextCommands.clear()


@VisiData.api
def moveToReplayContext(vd, r, vs):
        'set the sheet/row/col to the values in the replay row'
        vs.ensureLoaded()
        vd.sync()
        vd.clearCaches()

        if r.row not in [None, '']:
            vs.moveToRow(r.row) or vd.error(f'no "{r.row}" row on {vs}')

        if r.col not in [None, '']:
            vs.moveToCol(r.col) or vd.error(f'no "{r.col}" column on {vs}')


@VisiData.api
def replayOne(vd, r):
        'Replay the command in one given row.'
        vd.currentReplayRow = r
        longname = getattr(r, 'longname', None)
        if longname is None and getattr(r, 'keystrokes', None) is None:
            vd.fail('failed to find command to replay')

        if r.sheet and longname not in ['set-option', 'unset-option']:
            vs = vd.getSheet(r.sheet) or vd.error('no sheet named %s' % r.sheet)
        else:
            vs = None

        if longname in ['set-option', 'unset-option']:
            try:
                context = vs if r.sheet and vs else vd
                option_scope = r.sheet or r.col or 'global'
                if option_scope == 'override': option_scope = 'global' # override is deprecated, is now global
                if longname == 'set-option':
                    context.options.set(r.row, r.input, option_scope)
                else:
                    context.options.unset(r.row, option_scope)

                escaped = False
            except Exception as e:
                vd.exceptionCaught(e)
                escaped = True
        else:
            vs = vs or vd.activeSheet
            if vs:
                if vs in vd.sheets:  # if already on sheet stack, push to top
                    vd.push(vs)
            else:
                vs = vd.cmdlog

            try:
                vd.moveToReplayContext(r, vs)
                if r.comment:
                    vd.status(r.comment)

                # <=v1.2 used keystrokes in longname column; getCommand fetches both
                escaped = vs.execCommand(longname if longname else r.keystrokes, keystrokes=r.keystrokes)
            except Exception as e:
                vd.exceptionCaught(e)
                escaped = True

        vd.currentReplayRow = None

        if escaped:  # escape during replay aborts replay
            vd.warning('replay aborted during %s' % (longname or r.keystrokes))
        return escaped


@VisiData.api
class DisableAsync:
    def __enter__(self):
        vd.execAsync = vd.execSync

    def __exit__(self, exc_type, exc_val, tb):
        vd.execAsync = lambda *args, vd=vd, **kwargs: visidata.VisiData.execAsync(vd, *args, **kwargs)


@VisiData.api
def replay_sync(vd, cmdlog):
    'Replay all commands in *cmdlog*.'
    with vd.DisableAsync():
        vd.sync()  #2352 let cmdlog finish loading
        cmdlog.cursorRowIndex = 0
        # save current replay, for cmdlogs that replay other cmdlogs, such as a macro executing another macro
        prev_replay = vd.currentReplay
        prev_replay_row = vd.currentReplayRow
        vd.currentReplay = cmdlog

        with Progress(total=len(cmdlog.rows)) as prog:
            while cmdlog.cursorRowIndex < len(cmdlog.rows):
                if vd.currentReplay is None:
                    vd.status('replay canceled')
                    return

                vd.statuses.clear()
                try:
                    if vd.replayOne(cmdlog.cursorRow):
                        vd.replay_cancel()
                        return True
                except Exception as e:
                    vd.replay_cancel()
                    vd.exceptionCaught(e)
                    vd.status('replay canceled')
                    return True

                cmdlog.cursorRowIndex += 1
                prog.addProgress(1)

                if vd.activeSheet:
                    vd.activeSheet.ensureLoaded()

        vd.status('replay complete')
        vd.currentReplay = prev_replay
        vd.currentReplayRow = prev_replay_row


@VisiData.api
def replay(vd, cmdlog):
    'Inject commands into live execution with interface.'
    vd.push(cmdlog)
    vd._nextCommands.extend(cmdlog.rows)


@VisiData.api
def getLastArgs(vd):
        'Get user input for the currently playing command.'
        if vd.currentReplayRow:
            return vd.currentReplayRow.input
        return None


@VisiData.api
def setLastArgs(vd, args):
        'Set user input on last command, if not already set.'
        # only set if not already set (second input usually confirmation)
        if vd.activeCommand:
            if not vd.activeCommand.input:
                vd.activeCommand.input = args


@VisiData.property
def replayStatus(vd):
    if vd.macroMode:
        return f'|[:error] {len(vd.macroMode)} {vd.options.disp_replay_record} [:]'

    if vd._nextCommands:
        return f'|[:status_replay] {len(vd._nextCommands)} {vd.options.disp_replay_play} [:]'

    return ''


@BaseSheet.property
def cmdlog(sheet):
    rows = sheet.cmdlog_sheet.rows
    if isinstance(sheet.source, BaseSheet):
        rows = sheet.source.cmdlog.rows + rows
    return CommandLogJsonl(sheet.name+'_cmdlog', source=sheet, rows=rows)


@BaseSheet.lazy_property
def cmdlog_sheet(sheet):
    c = CommandLogJsonl(sheet.name+'_cmdlog', source=sheet, rows=[])
    # copy over all existing globally set options
    # you only need to do this for the first BaseSheet in a tree
    if not isinstance(sheet.source, BaseSheet):
        for r in vd.cmdlog.rows:
            if r.sheet == 'global' and (r.longname == 'set-option') or (r.longname == 'unset-option'):
                c.addRow(r)
    return c


@BaseSheet.property
def shortcut(self):
    if self._shortcut:
        return self._shortcut
    try:
        return str(vd.allSheets.index(self)+1)
    except ValueError:
        pass

    try:
        return self.cmdlog_sheet.rows[0].keystrokes or ''  #2293
    except Exception:
        pass

    return ''


@VisiData.property
def cmdlog(vd):
    if not vd._cmdlog:
        vd._cmdlog = CommandLogJsonl('cmdlog', rows=[])  # no reload
        vd._cmdlog.resetCols()
        vd.beforeExecHooks.append(vd._cmdlog.beforeExecHook)
    return vd._cmdlog

@VisiData.property
def modifyCommand(vd):
    if vd.activeCommand and vd.isLoggableCommand(vd.activeCommand):
        return vd.activeCommand
    if not vd.cmdlog.rows:
        return None
    return vd.cmdlog.rows[-1]


@CommandLogJsonl.api
@asyncthread
def repeat_for_n(cmdlog, r, n=1):
    r.sheet = r.row = r.col = ""
    for i in range(n):
        vd.replayOne(r)

@CommandLogJsonl.api
@asyncthread
def repeat_for_selected(cmdlog, r):
    r.sheet = r.row = r.col = ""

    for idx, r in enumerate(vd.sheet.rows):
        if vd.sheet.isSelected(r):
            vd.sheet.cursorRowIndex = idx
            vd.replayOne(r)


BaseSheet.init('_shortcut')


globalCommand('gD', 'cmdlog-all', 'vd.push(vd.cmdlog)', 'open global CommandLog for all commands executed in current session')
globalCommand('D', 'cmdlog-sheet', 'vd.push(sheet.cmdlog)', "open current sheet's CommandLog with all other loose ends removed; includes commands from parent sheets")
globalCommand('zD', 'cmdlog-sheet-only', 'vd.push(sheet.cmdlog_sheet)', 'open CommandLog for current sheet with commands from parent sheets removed')
BaseSheet.addCommand('^D', 'save-cmdlog', 'saveSheets(inputPath("save cmdlog to: ", value=fnSuffix(name)), vd.cmdlog)', 'save CommandLog to filename.vdj file')
BaseSheet.bindkey('^N', 'no-op')
BaseSheet.addCommand('^K', 'replay-stop', 'vd.replay_cancel(); vd.warning("replay canceled")', 'cancel current replay')

globalCommand(None, 'show-status', 'status(input("status: "))', 'show given message on status line')
globalCommand('^V', 'show-version', 'status(__version_info__);', 'Show version and copyright information on status line')
globalCommand('z^V', 'check-version', 'checkVersion(input("require version: ", value=__version_info__))', 'check VisiData version against given version')

CommandLog.addCommand('x', 'replay-row', 'vd.replayOne(cursorRow); status("replayed one row")', 'replay command in current row')
CommandLog.addCommand('gx', 'replay-all', 'vd.replay(sheet)', 'replay contents of entire CommandLog')

CommandLogJsonl.addCommand('x', 'replay-row', 'vd.replayOne(cursorRow); status("replayed one row")', 'replay command in current row')
CommandLogJsonl.addCommand('gx', 'replay-all', 'vd.replay(sheet)', 'replay contents of entire CommandLog')

CommandLog.options.json_sort_keys = False
CommandLog.options.encoding = 'utf-8'
CommandLogJsonl.options.json_sort_keys = False
CommandLogJsonl.options.regex_skip = r'^(//|#).*'

vd.addGlobals(CommandLogBase=CommandLogBase, CommandLogRow=CommandLogRow)

vd.addMenuItems('''
            View > Command log > this sheet > cmdlog-sheet
    View > Command log > this sheet only > cmdlog-sheet-only
    View > Command log > all commands > cmdlog-all
    System > Execute longname > exec-longname
    Help > Version > show-version
''')



'''
# A Guide to VisiData Guides
Each guide shows you how to use a particular feature in VisiData. Gray guides have not been written yet. We love contributions: [:onclick https://visidata.org/docs/api/guides]https://visidata.org/docs/api/guides[/].

- [:keystrokes]Up/Down[/] to move the row cursor
- [:keystrokes]Enter[/] to view a topic
'''
import re

from visidata import vd, BaseSheet, Sheet, ItemColumn, Column, VisiData, ENTER, RowColorizer, AttrDict, MissingAttrFormatter
from visidata import wraptext, Path, CellColorizer
import visidata

guides_list = '''
GuideIndex ("A Guide to VisiData Guides (you are here)")
HelpGuide ("Where to Start and How to Quit")  # manpage; ask for patreon
MenuGuide ("The VisiData Menu System")
CommandsSheet ("How to find the command you want run")

#  real barebones basics
MovementGuide ("Movement and search")
InputGuide ("Input keystrokes")
SortGuide ("Sorting")
ColumnsGuide ("Resize, hide, and rename columns")
TypesSheet ("Column types")
CommandLog  ("Undo and Replay")

#  rev this thing up

SelectionGuide ("Selecting and filtering") # stu|, and variants; filtering with dup-sheet; g prefix often refers to selected rowset
SheetsSheet  ("The Sheet Stack")
ColumnsSheet ("Columns: the only way to fly")
StatusesSheet ("Revisit old status messages")
SidebarSheet ("Dive into the sidebar")
SaversGuide ("Saving Data")  # talk about options.overwrite + ro here

ErrorsSheet ("What was that error?")
ModifyGuide ("Adding, Editing, Deleting Rows")

# the varieties of data experience

SlideGuide ("Sliding rows and columns around")
ExprGuide ("Compute Python over every row")
JoinGuide ("Joining multiple sheets together")
DescribeSheet ("Basic Statistics (min/max/mode/median/mean)")
AggregatorsSheet ("Aggregations like sum, mean, and distinct")
FrequencyTable ("Frequency Tables are how you GROUP BY")
PivotGuide ("Pivot Tables are just Frequency Tables with more columns")
MeltGuide ("Melt is just Unpivot")
JsonSheet ("Some special features for JSON") # with expand/contract, unfurl
RegexGuide ("Matching and Transforming Strings with Regex")
GraphSheet ("Basic scatterplots and other graphs")
WindowFunctionGuide ("Perform operations on groups of rows")

# for the frequent user
OptionsSheet ("Options and Settings")
ClipboardGuide ("Copy and Paste Data via the Clipboard")
DirSheet ("Browsing the local filesystem")
FormatsSheet ("What can you open with VisiData?")
SplitpaneGuide ("Split VisiData into two panes")
ThemesSheet ("Change Interface Theme")
ColorSheet ("See available colors")
MacrosSheet ("Recording macros")
MemorySheet ("Making note of certain values")

# Specific use cases

XsvGuide ("CSV/TSV and other text-delimited formats")
GrepSheet ("Load output of grep-like tools")

# advanced usage and developers

ThreadsSheet ("Threads past and present")
DeveloperGuide ("Inspecting internal Python objects")

#  appendices

InputEditorGuide ("Using the builtin line editor")
'''

vd.guides = {}  # name -> guidecls

@VisiData.api
def addGuide(vd, name):
    guideSource = Path(vd.pkg_resources_files(visidata)/f'guides/{name}.md')
    if guideSource.exists():
        vd.guides[name] = GuideSheet(name, source=guideSource)

@VisiData.api
class GuideIndex(Sheet):
    guide = __doc__

    rowtype = 'guides' # rowdef: list(guide number, guide name, topic description, points, max_points)
    columns = [
        ItemColumn('n', 0, width=0, type=int),
        ItemColumn('name', 1, width=0),
        ItemColumn('topic', 2, width=60),
    ]
    colorizers = [
            RowColorizer(7, 'color_guide_unwritten', lambda s,c,r,v: r and r[1] not in vd.guides)
            ]
    def iterload(self):
        i = 0
        for line in guides_list.splitlines():
            m = re.search(r'(\w+?) \("(.*)"\)', line)
            if m:
                guidename, description = list(m.groups())
                vd.addGuide(guidename)
                yield [i, guidename, description]
                i += 1

    def openRow(self, row):
        name = row[1]
        return vd.getGuide(name)

class OptionHelpGetter:
    'For easy and consistent formatting in sidebars and helpstrings, use {help.options.opt_name}.'
    def __getattr__(self, optname):
        opt = vd.options._get(optname, 'default')
        return f'[:onclick options-sheet {optname}][:longname_guide]{optname}[/][/]: {opt.helpstr} (default: {opt.value})'


class CommandHelpGetter:
    'For easy and consistent formatting in sidebars and helpstrings, use {help.commands.long_name}.'
    def __init__(self, cls):
        self.cls = cls
        self.helpsheet = vd.HelpSheet()
        vd.sync(self.helpsheet.ensureLoaded())

    def __getattr__(self, k):
        return self.__getitem__(k)

    def __getitem__(self, k):
        longname = k.replace('_', '-')
        binding = self.helpsheet.revbinds.get(longname, [None])[0] or '<unbound>'
        # cmddict has a SheetClass associated with each command
        # go through all the parents of the Sheet type, to look for the command
        for cls in self.cls.superclasses():
            cmd = self.helpsheet.cmddict.get((cls.__name__, longname), None)
            if cmd:
                break
        if not cmd:
            return ''
        if 'input' in cmd.execstr.lower():
            inputtype = 'input'
            m = re.search(r'type="(\w*)"', cmd.execstr, re.IGNORECASE)
            if not m:
                m = re.search(r'input(\w*)\("', cmd.execstr, re.IGNORECASE)
            if m:
                inputtype = m.groups()[0].lower() or 'input'
                binding += f' <{inputtype}>'

        helpstr = cmd.helpstr
        return f'[:code]{binding}[/] ([:longname_guide]{longname}[/]) to {helpstr}'


class GuideSheet(Sheet):
    rowtype = 'lines'
    filetype = 'guide'
    columns = [
            ItemColumn('linenum', 0, type=int, width=0),
            ItemColumn('guide', 1, width=80, displayer='full'),
            ]
    precious = False
    colorizers = [CellColorizer(4, 'color_sidebar', lambda s,c,r,v: True)]

    def iterload(self):
        self.metadata = AttrDict(sheettype='Sheet')
        text = self.source.open(mode='r').read()
        winWidth = 78

        # parsing front matter
        sections = text.split('---\n', maxsplit=2)
        for section in sections[:-1]:
            for config in section.splitlines():
                config = config.strip()
                if config:
                    try:
                        key, val = config.split(': ', maxsplit=1)
                    except ValueError:
                        vd.fail('incorrect front matter syntax')
                    self.metadata[key] = val

        # formatting text
        helper = AttrDict(commands=CommandHelpGetter(vd.getGlobals()[self.metadata.sheettype]),
                          options=OptionHelpGetter())
        guidetext = MissingAttrFormatter().format(sections[-1], help=helper, vd=vd)

        # parsing guide
        for startingLine, text in enumerate(guidetext.splitlines()):
            text = text.strip()
            if text:
                for i, (L, _) in enumerate(wraptext(str(text), width=winWidth)):
                    yield [startingLine+i+1, L]
            else:
                yield [startingLine+1, text]


@VisiData.api
def getGuide(vd, name): # -> GuideSheet()
    if name in vd.guides:
        return vd.guides[name]
    vd.warning(f'no guide named {name}')

BaseSheet.addCommand('', 'open-guide-index', 'vd.push(GuideIndex("VisiData_Guide"))', 'open VisiData guides table of contents')

@VisiData.api
def inputKeys(vd, prompt):
    return vd.input(prompt, help=f'''
                # Input Keystrokes
                - Press `Ctrl+N` and then press another keystroke to spell out that keystroke.
                - Press `Ctrl+C` to cancel the input.
                - Press `Enter` to accept the input.
            ''')

@BaseSheet.api
def getCommandInfo(sheet, keys):
    if not keys:
        return ''
    cmd = sheet.getCommand(keys)
    if cmd:
        return CommandHelpGetter(type(sheet))[cmd.longname]
    else:
        vd.warning(f'no command bound to {keys} on {sheet}')

GuideSheet.options.color_current_row = "underline"

vd.addCommand('', 'show-command-info', 'status(getCommandInfo(inputKeys("get command for keystrokes: ")))', 'show longname and helpstring for keybinding')

vd.addMenuItems('''
        Help > VisiData Feature Guides > open-guide-index
''')

vd.optalias('guides', 'preplay', 'open-guide-index')

vd.addGlobals({"CommandHelpGetter": CommandHelpGetter, "OptionHelpGetter": OptionHelpGetter})



from visidata import VisiData, Column, getitemdef, PythonSheet, asyncthread, vd


# requires (deb): libbz2-dev libcurl4-openssl-dev liblzma-dev

@VisiData.api
def open_vcf(vd, p):
    return VcfSheet(p.base_stem, source=p)

def unbox(col, row):
    v = getitemdef(row, col.expr)
    if not v:
        return None
    if len(v) == 1:
        return v[0].value
    return v


class VcfSheet(PythonSheet):
    rowtype = 'cards'
    @asyncthread
    def reload(self):
        vobject = vd.importExternal('vobject')
        self.rows = []
        self.columns = []

        addedCols = set()
        lines = []
        with self.open_text_source() as fp:
            for line in fp:
                lines.append(line)
                if line.startswith('END:'):
                    row = vobject.readOne('\n'.join(lines))
                    for k, v in row.contents.items():
                        if v and str(v[0].value).startswith('(None)'):
                            continue
                        if not k in addedCols:
                            addedCols.add(k)
                            self.addColumn(Column(k, expr=k, getter=unbox))
                    self.addRow(row.contents)
                    lines = []



from visidata import vd, IndexSheet, VisiData

'Requires visidata/deps/pyxlsb fork'

@VisiData.api
def guess_xls(vd, p):
    if p.open_bytes().read(16).startswith(b'\xD0\xCF\x11\xE0\xA1\xB1\x1A\xE1'):
        return dict(filetype='xlsb', _likelihood=10)


@VisiData.api
def open_xlsb(vd, p):
    return XlsbIndex(p.base_stem, source=p)


class XlsbIndex(IndexSheet):
    def iterload(self):
        vd.importExternal('pyxlsb', '-e git+https://github.com/saulpw/pyxlsb.git@visidata#egg=pyxlsb')
        from pyxlsb import open_workbook

        wb = open_workbook(str(self.source))
        for name in wb.sheets:
            yield wb.get_sheet(name, True)



import os
import contextlib
import itertools
import collections
import math
import time

from visidata import vd, asyncthread, options, Progress, ColumnItem, SequenceSheet, Sheet, VisiData
from visidata import namedlist, filesize

vd.option('delimiter', '\t', 'field delimiter to use for tsv/usv filetype', replay=True)
vd.option('row_delimiter', '\n', 'row delimiter to use for tsv/usv filetype', replay=True)
vd.option('tsv_safe_newline', '\u001e', 'replacement for newline character when saving to tsv', replay=True)
vd.option('tsv_safe_tab', '\u001f', 'replacement for tab character when saving to tsv', replay=True)


@VisiData.api
def open_tsv(vd, p):
    return TsvSheet(p.base_stem, source=p)


def adaptive_bufferer(fp, max_buffer_size=65536):
    """Loading e.g. tsv files goes faster with a large buffer. But when the input stream
    is slow (e.g. 1 byte/second) and the buffer size is large, it can take a long time until
    the buffer is filled. Only when the buffer is filled (or the input stream is finished)
    you can see the data visualized in visidata. That's why we use an adaptive buffer.
    For fast input streams, the buffer becomes large, for slow input streams, the buffer stays
    small"""
    buffer_size = 8
    processed_buffer_size = 0
    previous_start_time = time.time()
    while True:
        next_chunk = fp.read(max(buffer_size, 1))
        if not next_chunk:
            break

        yield next_chunk

        processed_buffer_size += len(next_chunk)

        current_time = time.time()
        current_delta = current_time - previous_start_time

        if current_delta < 1:
            # if it takes less than one second to fill the buffer, double the size of the buffer
            buffer_size = min(buffer_size * 2, max_buffer_size)
        else:
            # if it takes longer than one second, decrease the buffer size so it takes about
            # 1 second to fill it
            previous_start_time = current_time
            buffer_size = math.ceil(min(processed_buffer_size / current_delta, max_buffer_size))
            processed_buffer_size = 0

def splitter(stream, delim='\n'):
    'Generates one line/row/record at a time from stream, separated by delim'

    buf = type(delim)()

    for chunk in stream:
        buf += chunk

        *rows, buf = buf.split(delim)
        yield from rows

    buf = buf.rstrip(delim)  # trim empty trailing lines
    if buf:
        yield from buf.rstrip(delim).split(delim)


# rowdef: list
class TsvSheet(SequenceSheet):
    delimiter = ''
    row_delimiter = ''

    def iterload(self):
        delim = self.delimiter or self.options.delimiter
        rowdelim = self.row_delimiter or self.options.row_delimiter
        if delim == '':
            vd.warning("using '\\x00' as field delimiter")
            delim = '\x00'  #2272
            self.options.regex_skip = ''
        if rowdelim == '':
            vd.warning("using '\\x00' as row delimiter")
            rowdelim = '\x00'
            self.options.regex_skip = ''
        if delim == rowdelim:
            vd.fail('field delimiter and row delimiter cannot be the same')

        with self.open_text_source() as fp:
                regex_skip = getattr(fp, '_regex_skip', None)
                for line in splitter(adaptive_bufferer(fp), rowdelim):
                    if not line or (regex_skip and regex_skip.match(line)):
                        continue

                    row = line.split(delim)

                    if len(row) < self.nVisibleCols:
                        # extend rows that are missing entries
                        row.extend([None]*(self.nVisibleCols-len(row)))

                    yield row


@VisiData.api
def save_tsv(vd, p, vs, delimiter='', row_delimiter=''):
    'Write sheet to file `fn` as TSV.'
    unitsep = delimiter or vs.options.delimiter
    rowsep = row_delimiter or vs.options.row_delimiter
    if unitsep == '':
        vd.warning("saving with '\\x00' as field delimiter")
        unitsep = '\x00'
    if rowsep == '':
        vd.warning("saving with '\\x00' as row delimiter")
        rowsep = '\x00'
    if unitsep == rowsep:
        vd.fail('field delimiter and row delimiter cannot be the same')
    trdict = vs.safe_trdict()

    with p.open(mode='w', encoding=vs.options.save_encoding) as fp:
        colhdr = unitsep.join(col.name.translate(trdict) for col in vs.visibleCols) + rowsep
        fp.write(colhdr)

        for dispvals in vs.iterdispvals(format=True):
            fp.write(unitsep.join(dispvals.values()))
            fp.write(rowsep)


@Sheet.api
def append_tsv_row(vs, row):
    'Append `row` to vs.source, creating file with correct headers if necessary. For internal use only.'
    if not vs.source.exists():
        with contextlib.suppress(FileExistsError):
            parentdir = vs.source.parent
            if parentdir:
                os.makedirs(parentdir)

        # Write tsv header for Sheet `vs` to Path `p`
        trdict = vs.safe_trdict()
        unitsep = options.delimiter

        with vs.source.open(mode='w') as fp:
            colhdr = unitsep.join(col.name.translate(trdict) for col in vs.visibleCols) + vs.options.row_delimiter
            if colhdr.strip():  # is anything but whitespace
                fp.write(colhdr)

    newrow = ''

    contents = vs.source.open(mode='r').read()
    if not contents.endswith('\n'):  #1569
        newrow += '\n'

    newrow += '\t'.join(col.getDisplayValue(row) for col in vs.visibleCols) + '\n'

    with vs.source.open(mode='a') as fp:
        fp.write(newrow)


vd.addGlobals({
    'TsvSheet': TsvSheet,
})



import os
import os.path
import sys
import io
import codecs
import pathlib
from urllib.parse import urlparse, urlunparse
from functools import wraps, lru_cache

from visidata import vd
from visidata import VisiData, Progress

vd.help_encoding = '''Common Encodings:

- `utf-8`: Unicode (ASCII compatible, most common)
- `utf-8-sig`: Unicode as above, but saves/skips leading BOM
- `ascii`: 7-bit ASCII
- `latin1`: also known as `iso-8859-1`
- `cp437`: original IBM PC character set
- `shift_jis`: Japanese

See [:onclick https://docs.python.org/3/library/codecs.html#standard-encodings]https://docs.python.org/3/library/codecs.html#standard-encodings[/]
'''

vd.help_encoding_errors = '''Encoding Error Handlers:

- `strict`: raise error
- `ignore`: discard
- `replace`: replacement marker
- `backslashreplace`: use "\\uxxxxxx"
- `surrogateescape`: use surrogate characters

See [:onclick https://docs.python.org/3/library/codecs.html#error-handlers]https://docs.python.org/3/library/codecs.html#error-handlers[/]
'''

vd.option('encoding', 'utf-8-sig', 'encoding passed to codecs.open when reading a file', replay=True, help=vd.help_encoding)
vd.option('encoding_errors', 'surrogateescape', 'encoding_errors passed to codecs.open', replay=True, help=vd.help_encoding_errors)

@VisiData.api
def pkg_resources_files(vd, package):
    '''
    Returns a Traversable object (Path-like), based on the location of the package.
    importlib.resources.files exists in Python >= 3.9; use importlib_resources for the rest.
    '''
    try:
        from importlib.resources import files
    except ImportError: #1968
        from importlib_resources import files
    return files(package)

@lru_cache()
def vstat(path, force=False):
    try:
        return os.stat(path)
    except Exception as e:
        return None

def filesize(path):
    if hasattr(path, 'filesize') and path.filesize is not None:
        return path.filesize
    if hasattr(path, 'is_url') and path.is_url():
        return 0
    if hasattr(path, 'has_fp') and path.has_fp():
        return 0
    st = path.stat() # vstat(path)
    return st and st.st_size

def modtime(path):
    st = path.stat()
    return st and st.st_mtime


# from https://stackoverflow.com/questions/55889474/convert-io-stringio-to-io-bytesio
class BytesIOWrapper(io.BufferedReader):
    """Wrap a buffered bytes stream over TextIOBase string stream."""

    def __init__(self, text_io_buffer, encoding=None, errors=None, **kwargs):
        super(BytesIOWrapper, self).__init__(text_io_buffer, **kwargs)
        self.encoding = encoding or text_io_buffer.encoding or vd.options.encoding
        self.errors = errors or text_io_buffer.errors or vd.options.encoding_errors

    def _encoding_call(self, method_name, *args, **kwargs):
        raw_method = getattr(self.raw, method_name)
        val = raw_method(*args, **kwargs)
        return val.encode(self.encoding, errors=self.errors)

    def read(self, size=-1):
        return self._encoding_call('read', size)

    def read1(self, size=-1):
        return self._encoding_call('read1', size)

    def peek(self, size=-1):
        return self._encoding_call('peek', size)


class FileProgress:
    'Open file in binary mode and track read() progress.'
    def __init__(self, path, fp, mode='r', **kwargs):
        self.path = path
        self.fp = fp
        self.prog = None
        if 'r' in mode:
            gerund = 'reading'
            self.prog = Progress(gerund=gerund, total=filesize(path))
        elif 'w' in mode:
            gerund = 'writing'
            self.prog = Progress(gerund=gerund)
        else:
            gerund = 'nothing'

        # track Progress on original fp
        self.fp_orig_read = self.fp.read
        self.fp_orig_readline = self.fp.readline
        self.fp_orig_close = self.fp.close

        self.fp.read = self.read
        self.fp.close = self.close

        if self.prog:
            self.prog.__enter__()

    def close(self, *args, **kwargs):
        if self.prog:
            self.prog.__exit__(None, None, None)
            self.prog = None
        return self.fp_orig_close(*args, **kwargs)

    def read(self, size=-1):
        r = self.fp_orig_read(size)
        if self.prog:
            if r:
                self.prog.addProgress(len(r))
        return r

    def readline(self, size=-1):
        r = self.fp_orig_readline(size)
        if self.prog:
            self.prog.addProgress(len(r))
        return r

    def __getattr__(self, k):
        return getattr(self.fp, k)

    def __enter__(self):
        self.fp.__enter__()
        return self

    def __next__(self):
        r = next(self.fp)
        self.prog.addProgress(len(r))
        return r

    def __iter__(self):
        if not self.prog:
            yield from self.fp
        else:
            for line in self.fp:
                self.prog.addProgress(len(line))
                yield line

    def __exit__(self, type, value, tb):
        return self.fp.__exit__(type, value, tb)


class Path(os.PathLike):
    'File and path-handling class, modeled on `pathlib.Path`.'
    def __init__(self, given, fp=None, fptext=None, lines=None, filesize=None):
        # Resolve pathname shell variables and ~userdir
        self.given = os.path.expandvars(os.path.expanduser(str(given)))
        self.fptext = fptext
        self.fp = fp
        self.lines = lines or []  # shared among all RepeatFile instances
        self.filesize = filesize
        self.rfile = None

    @property
    def name(self):
        'Filename without any extensions.  Not the same as pathlib.Path.'
        return self.base_stem

    @property
    def given(self):
        'The path as given to the constructor.'
        return self._given

    @given.setter
    def given(self, given):
        self._given = given
        if isinstance(given, os.PathLike):
            self._path = given
        else:
            self._path = pathlib.Path(given)

        self.ext = self.suffix[1:]
        if self.suffix:  #1450  don't make this a oneliner; [:-0] doesn't work
            self.base_stem = self._path.name[:-len(self.suffix)]
        elif self._given == '.':  #1768
            self.base_stem = self._path.absolute().name
        else:
            self.base_stem = self._path.name

        # check if file is compressed
        if self.suffix in ['.gz', '.bz2', '.xz', '.lzma', '.zst']:
            self.compression = self.ext
            uncompressedpath = Path(self.given[:-len(self.suffix)])  # strip suffix
            self.base_stem = uncompressedpath.base_stem
            self.ext = uncompressedpath.ext
        else:
            self.compression = None

    def __getattr__(self, k):
        if hasattr(self.__dict__, k):
            r = getattr(self.__dict__, k)
        else:
            if self.__dict__.get('_path', None) is not None:
                r = getattr(self._path, k)
            else:
                raise AttributeError(k)
        if isinstance(r, pathlib.Path):
            return Path(r)
        return r

    def __fspath__(self):
        return self._path.__fspath__()

    def __lt__(self, a):
        if isinstance(a, Path):
            return self._path.__lt__(a._path)
        return self._path.__lt__(a)

    def __truediv__(self, a):
        return Path(self._path.__truediv__(a))

    def has_fp(self):
        'Return True if this is a virtual Path to an already open file.'
        return bool(self.fp or self.fptext)

    def open(self, mode='rt', encoding=None, encoding_errors=None, newline=None):
        if 'b' in mode:
            return self.open_bytes(mode)

        return self.open_text(mode=mode, encoding=encoding, encoding_errors=encoding_errors, newline=newline)

    def open_bytes(self, mode='rb'):
        'Open the file pointed by this path and return a file object in binary mode.'
        if self.rfile:
            raise ValueError('a RepeatFile holds text and cannot be reopened in binary mode')

        if 'b' not in mode:
            mode += 'b'

        if self.given == '-':
            if 'r' in mode:
                return os.fdopen(vd._stdin.fileno(), 'rb')
            elif 'w' in mode or 'a' in mode:
                # convert 'a' to 'w' for stdout: https://bugs.python.org/issue27805
                return os.dup(vd._stdout.fileno())
            else:
                vd.error('invalid mode "%s" for Path.open()' % mode)
                return sys.stderr

        return self._open(mode=mode)

    def open_text(self, mode='rt', encoding=None, encoding_errors=None, newline=None):
        'Open path in text mode, using options.encoding and options.encoding_errors.  Return open file-pointer or file-pointer-like.'
        # rfile makes a single-access fp reusable

        if 't' not in mode:
            mode += 't'

        if self.rfile:
            return self.rfile.reopen()

        if self.fp and not self.fptext:
            self.fptext = codecs.iterdecode(self.fp,
                                            encoding=encoding or vd.options.encoding,
                                            errors=encoding_errors or vd.options.encoding_errors)

        if self.fptext:
            self.rfile = RepeatFile(self.fptext)
            return self.rfile

        if self.given == '-':
            if 'r' in mode:
                return vd._stdin
            elif 'w' in mode or 'a' in mode:
                # convert 'a' to 'w' for stdout: https://bugs.python.org/issue27805
                return open(os.dup(vd._stdout.fileno()), 'wt')
            else:
                vd.error('invalid mode "%s" for Path.open()' % mode)
                return sys.stderr

        return self._open(mode=mode, encoding=encoding or vd.options.encoding, errors=vd.options.encoding_errors, newline=newline)

    @wraps(pathlib.Path.read_text)
    def read_text(self, *args, **kwargs):
        'Open the file in text mode and return its entire decoded contents.'
        if 'encoding' not in kwargs:
            kwargs['encoding'] = vd.options.encoding
        if 'errors' not in kwargs:
            kwargs['errors'] = kwargs.get('encoding_errors', vd.options.encoding_errors)

        if self.lines:
            return RepeatFile(self.lines).read()
        elif self.fp:
            return self.fp.read()
        elif self.fptext:
            return self.fptext.read()
        else:
            return self._path.read_text(*args, **kwargs)

    @wraps(pathlib.Path.open)
    def _open(self, *args, **kwargs):
        if self.fp:
            return FileProgress(self, fp=self.fp, **kwargs)

        if self.fptext:
            return FileProgress(self, fp=BytesIOWrapper(self.fptext), **kwargs)

        path = self

        if self.compression == 'gz':
            import gzip
            zopen = gzip.open
        elif self.compression == 'bz2':
            import bz2
            zopen = bz2.open
        elif self.compression in ['xz', 'lzma']:
            import lzma
            zopen = lzma.open
        elif self.compression == 'zst':
            zstandard = vd.importExternal('zstandard')
            zopen = zstandard.open
        else:
            return FileProgress(path, fp=self._path.open(*args, **kwargs), **kwargs)

        if 'w' in kwargs.get('mode', ''):
            #1159 FileProgress on the outside to close properly when writing
            return FileProgress(path, fp=zopen(path, **kwargs), **kwargs)

        #1255 FileProgress on the inside to track uncompressed bytes when reading
        return zopen(FileProgress(path, fp=open(path, mode='rb'), **kwargs), **kwargs)

    def __iter__(self):
        with Progress(total=filesize(self)) as prog:
            with self.open(encoding=vd.options.encoding) as fd:
                for i, line in enumerate(fd):
                    prog.addProgress(len(line))
                    yield line.rstrip('\n')

    def read_bytes(self):
        'Return the entire binary contents of the pointed-to file as a bytes object.'
        with self.open(mode='rb') as fp:
            return fp.read()

    @wraps(pathlib.Path.is_fifo)
    def is_fifo(self):
        'Return True if the path is a fifo.'
        return self._path.is_fifo()

    def is_local(self):
        'Return True if self.filename refers to a file on the local disk.'
        return not bool(self.is_url()) and not bool(self.fp) and not bool(self.fptext)

    def is_url(self):
        'Return True if the given path appears to be a URL.'
        return '://' in self.given

    def __str__(self):
        if self.is_url():
            return self.given
        return str(self._path)

    @wraps(pathlib.Path.stat)
    @lru_cache()
    def stat(self, force=False):
        'Return Path.stat() if relevant.'
        try:
            if not self.is_url():
                return self._path.stat()
        except Exception as e:
            return None

    @wraps(pathlib.Path.exists)
    def exists(self):
        'Return True if the path can be opened.'
        if self.has_fp() or self.is_url():
            return True
        return self._path.exists()

    @property
    def scheme(self):
        'The URL scheme component, if path is a URL.'
        if self.is_url():
            return urlparse(self.given).scheme

    def with_name(self, name):
        'Return a sibling Path with *name* as a filename in the same directory.'
        if self.is_url():
            urlparts = list(urlparse(self.given))
            urlparts[2] = '/'.join(list(Path(urlparts[2]).parts[1:-1]) + [name])
            return Path(urlunparse(urlparts))
        else:
            return Path(self._from_parsed_parts(self._drv, self._root, list(self.parts[:-1]) + [name]))


class RepeatFile:
    '''Lazy file-like object that can be read and line-seeked more than once from memory.'''

    def __init__(self, iter_lines, lines=None):
        self.iter_lines = iter_lines
        self.lines = lines if lines is not None else []
        self.iter = RepeatFileIter(self)

    def __enter__(self):
        '''Returns a new independent file-like object, sharing the same line cache.'''
        return self.reopen()

    def __exit__(self, a,b,c):
        pass

    def reopen(self):
        'Return copy of file-like with internal iterator reset.'
        return RepeatFile(self.iter_lines, lines=self.lines)

    def read(self, n=None):
        '''Returns a string or bytes object. Unlike the standard read() function, when *n* is given, more than *n* characters/bytes can be returned, and often will.'''
        if n is None:
            n = 10**12  # some too huge number
        r = []
        size = 0
        output_type = str; eol = '\n'; joiner = ''
        while not r or size < n:
            try:
                s = next(self.iter)
                if not r and isinstance(s, bytes):
                    output_type = bytes; eol = b'\n'; joiner = b''
                assert isinstance(s, output_type), (s, output_type)
                r.append(s)
                r.append(eol)
                size += len(s) + len(eol)
            except StopIteration:
                break  # end of file
        return joiner.join(r)

    def write(self, s):
        return self.iter_lines.write(s)

    def tell(self):
        '''Tells the current position as an opaque line marker.'''
        return self.iter.nextIndex

    def seek(self, offset, whence=io.SEEK_SET):
        '''Seek to an already seen opaque line position marker only.'''
        if whence != io.SEEK_SET and offset != 0:
            if whence == io.SEEK_CUR:
                raise io.UnsupportedOperation("can't do nonzero cur-relative seeks")
            elif whence == io.SEEK_END:
                raise io.UnsupportedOperation("can't do nonzero end-relative seeks")
            else:
                raise ValueError('invalid whence (%s, should be %s, %s or %s)' % (whence, io.SEEK_SET, io.SEEK_CUR, io.SEEK_END))
        self.iter.nextIndex = offset

    def readline(self, size=-1):
        if size != -1:
            vd.error('RepeatFile does not support limited line length')
        try:
            return next(self.iter)
        except StopIteration:
            return ''

    def __iter__(self):
        return RepeatFileIter(self)

    def __next__(self):
        return next(self.iter)

    def exists(self):
        return True


class RepeatFileIter:
    def __init__(self, rf):
        self.rf = rf
        self.nextIndex = 0

    def __iter__(self):
        return RepeatFileIter(self.rf)

    def __next__(self):
        if self.nextIndex < len(self.rf.lines):
            r = self.rf.lines[self.nextIndex]
        elif self.rf.iter_lines:
            try:
                r = next(self.rf.iter_lines)
                self.rf.lines.append(r)
            except StopIteration:
                self.rf.iter_lines = None
                raise
        else:
            raise StopIteration()


        self.nextIndex += 1
        return r


vd.addGlobals(RepeatFile=RepeatFile,
              Path=Path,
              modtime=modtime,
              filesize=filesize,
              vstat=vstat)



from visidata import *


@VisiData.api
def new_top(vd, p):
    return vd.processes


class CPUStatsSheet(Sheet):
    rowtype='CPUs'  # rowdef = (count, perfect, times, freq) from psutil (see below)
    columns = [
        ColumnItem('cpu', 0, type=int, keycol=0),
        ColumnItem('cpu_pct', 1, type=float)
    ]
    def iterload(self):
        psutil = vd.importExternal('psutil')

        for i, k in enumerate(psutil.cpu_times()._fields):
            self.addColumn(SubColumnItem(2, ColumnItem(k+'_s', i, type=float, sheet=self)))

        for i, k in enumerate(psutil.cpu_freq()._fields):
            self.addColumn(SubColumnItem(3, ColumnItem(k+'_MHz', i, type=float, sheet=self)))

        for r in zip(range(psutil.cpu_count()),
                        psutil.cpu_percent(percpu=True),
                        psutil.cpu_times(percpu=True),
                        psutil.cpu_freq(percpu=True)):
            yield r



class MemStatsSheet(Sheet):
    rowtype = ''  # rowdef: (name, value)
    columns = [
        AttrColumn('t', type=date, fmtstr='%H:%M:%S'),
        AttrColumn('meminfo'),
        AttrColumn('virtmem'),
        AttrColumn('swapmem'),
    ]
    nKeys = 1

    def iterload(self):
        psutil = vd.importExternal('psutil')
        import time

        proc = psutil.Process()

        while True:
            yield AttrDict(t=time.time(),
                         virtmem=psutil.virtual_memory(),
                         swapmem=psutil.swap_memory(),
                         meminfo=proc.memory_info(),
                        )

            if self.nRows == 1:  # first time through
                for n in 'meminfo virtmem swapmem'.split():
                    c = self.column(n)
                    if not c.hidden:
                        c.expand([self.rows[0]])

            time.sleep(1)


class UsefulProcessesSheet(Sheet):
    columns = [
        Column('pid', type=int, getter=lambda c,r: r[0].pid),
        Column('name', getter=lambda c,r: r[0].name()),
        Column('status', getter=lambda c,r: r[0].status()),
        Column('cmdline', getter=lambda c,r: ' '.join(r[0].cmdline())),
        Column('user', getter=lambda c,r: r[0].username()),
        Column('real_uid', type=int, width=0, getter=lambda c,r: r[0].uids()[0]),
        Column('effective_uid', type=int, width=0, getter=lambda c,r: r[0].uids()[1]),
        Column('mem_uss', type=int, getter=lambda c,r: r[1].mem_uss),
        Column('user_time_s', type=float, getter=lambda c,r: r[0].cpu_times()[0]),
        Column('system_time_s', type=float, getter=lambda c,r: r[0].cpu_times()[1]),
    ]
    def iterload(self):
        psutil = vd.importExternal('psutil')
        for pr in psutil.process_iter():
            yield (pr, pr.memory_full_info())


class ProcessesSheet(Sheet):
    columns = [
        Column('pid', type=int, getter=lambda c,r: r.pid),
        Column('name', getter=lambda c,r: r.name()),
        Column('status', getter=lambda c,r: r.status()),
        Column('parent_pid', type=int, getter=lambda c,r: r.ppid()),
        Column('exe', getter=lambda c,r: r.exe()),
        Column('cmdline', getter=lambda c,r: ' '.join(r.cmdline())),
        Column('cwd', getter=lambda c,r: r.cwd()),
        Column('user', getter=lambda c,r: r.username()),

        Column('real_uid', type=int, width=0, getter=lambda c,r: r.uids()[0]),
        Column('effective_uid', type=int, width=0, getter=lambda c,r: r.uids()[1]),
        Column('saved_uid', type=int, width=0, getter=lambda c,r: r.uids()[2]),

        Column('real_gid', type=int, width=0, getter=lambda c,r: r.gids()[0]),
        Column('effective_gid', type=int, width=0, getter=lambda c,r: r.gids()[1]),
        Column('saved_gid', type=int, width=0, getter=lambda c,r: r.gids()[2]),

        Column('create_time', type=date, getter=lambda c,r: r.create_time()),
        Column('cpu_num', type=int, getter=lambda c,r: r.cpu_num()),
        Column('cpu_percent', type=float, getter=lambda c,r: r.cpu_percent()),

        Column('tty', getter=lambda c,r: r.terminal()),
        Column('nice', getter=lambda c,r: r.nice()),
        Column('ioclass', getter=lambda c,r: r.ionice()[0]),
        Column('ionice', getter=lambda c,r: r.ionice()[1]),

        Column('user_time_s', type=float, getter=lambda c,r: r.cpu_times()[0]),
        Column('system_time_s', type=float, getter=lambda c,r: r.cpu_times()[1]),
        Column('children_user_time_s', type=float, width=0, getter=lambda c,r: r.cpu_times()[2]),
        Column('children_system_time_s', type=float, width=0, getter=lambda c,r: r.cpu_times()[3]),

        Column('read_ops', type=int, getter=lambda c,r: r.io_counters()[0]),
        Column('write_ops', type=int, getter=lambda c,r: r.io_counters()[1]),
        Column('read_bytes', type=int, getter=lambda c,r: r.io_counters()[2]),
        Column('write_bytes', type=int, getter=lambda c,r: r.io_counters()[3]),

        Column('voluntary_ctx_switches', type=int, getter=lambda c,r: r.num_ctx_switches()[0]),
        Column('involuntary_ctx_switches', type=int, getter=lambda c,r: r.num_ctx_switches()[1]),
        Column('num_fds', type=int, getter=lambda c,r: r.num_fds()),
        Column('num_threads', type=int, getter=lambda c,r: r.num_threads()),
    ]
    nKeys = 2

    def iterload(self):
        psutil = vd.importExternal('psutil')
#        self.columns = []
#        for c in ProcessesSheet.columns:
#            self.addColumn(c)

        for i,k in enumerate(psutil.Process().memory_full_info()._fields):
            self.addColumn(Column('mem_'+k, type=int, getter=lambda c,r,i=i: r.memory_full_info()[i], cache=True))
        # mem_uss is probably the most representative metric for determining how much memory is actually being used by a process. It represents the amount of memory that would be freed if the process was terminated right now.

        yield from psutil.process_iter()
#            try:
#                self.addRow((pr, pr.memory_full_info()))
#            except:
#                self.addRow((pr, None))

class RlimitsSheet(Sheet):
    columns = [
        ColumnItem('rlimit', 0),
        Column('soft', type=int, getter=lambda c,r: c.sheet.soft(r), setter=lambda c,r,v: c.sheet.set_soft(r, v)),
        Column('hard', type=int, getter=lambda c,r: c.sheet.hard(r), setter=lambda c,r,v: c.sheet.set_hard(r, v))
    ]

    def soft(self, r):
        return self.source.rlimit(r[1])[0]
    def hard(self, r):
        return self.source.rlimit(r[1])[1]
    def set_soft(self, r, v):
        self.source.rlimit(r[1], (v, self.hard(r)))
    def set_hard(self, r, v):
        self.source.rlimit(r[1], (self.soft(r), v))

    def iterload(self):
        psutil = vd.importExternal('psutil')
        for r in dir(psutil):
            if r.startswith('RLIMIT'):
                yield (r[7:], getattr(psutil, r))


@VisiData.lazy_property
def cpuStats(vd):
    return CPUStatsSheet('cpu_stats')

@VisiData.lazy_property
def memStats(vd):
    return MemStatsSheet('memory_stats')

@VisiData.lazy_property
def processes(vd):
    return ProcessesSheet('processes')


BaseSheet.addCommand('', 'open-cpustats', 'vd.push(vd.cpuStats)', 'open CPU usage stats' )
BaseSheet.addCommand('', 'open-memstats', 'vd.push(vd.memStats)', 'open memory usage stats')
BaseSheet.addCommand('', 'open-processes', 'vd.push(vd.processes)', 'open system process stats')

@VisiData.api
def chooseSignal(vd):
    import signal
    d = [{'key': attr[3:]} for attr in dir(signal) if attr.startswith('SIG') and not attr.startswith('SIG_')]
    return getattr(signal, 'SIG'+vd.chooseOne(d, type='signal'))

ProcessesSheet.addCommand('d', 'term-process', 'os.kill(cursorRow.pid, signal.SIGTERM)', 'send SIGTERM to process')
ProcessesSheet.addCommand('gd', 'term-selected', 'for r in someSelectedRows: os.kill(r.pid, signal.SIGTERM)', 'send SIGTERM to selected processes')
ProcessesSheet.addCommand('zd', 'kill-process', 'os.kill(cursorRow.pid, signal.SIGKILL)', 'send SIGKILL to process')
ProcessesSheet.addCommand('gzd', 'kill-selected', 'for r in someSelectedRows: os.kill(r.pid, signal.SIGKILL)', 'send SIGKILL to selected processes')

ProcessesSheet.addCommand('^K', 'signal-process', 'os.kill(cursorRow.pid, chooseSignal())', 'send chosen signal to process')
UsefulProcessesSheet.addCommand('^K', 'signal-selected', 'os.kill(cursorRow.pid, chooseSignal())', 'kill(2) send chosen signal to process')
UsefulProcessesSheet.addCommand('', 'open-rlimits', 'vd.push(RlimitsSheet(cursorRow.name() + "_rlimits", cursorRow))', 'push rlimits for this process')

vd.addMenuItem('System', 'Signal', 'current process', 'TERMinate', 'term-process')
vd.addMenuItem('System', 'Signal', 'current process', 'KILL', 'kill-process')
vd.addMenuItem('System', 'Signal', 'current process', 'choose signal', 'signal-process')
vd.addMenuItem('System', 'Signal', 'selected processes', 'TERMinate', 'term-selected')
vd.addMenuItem('System', 'Signal', 'selected processes', 'KILL', 'kill-selected')
vd.addMenuItem('System', 'Signal', 'selected processes', 'choose signal', 'signal-selected')
vd.addMenuItem('System', 'Stats', 'CPU', 'open-cpustats')
vd.addMenuItem('System', 'Stats', 'memory', 'open-memstats')
vd.addMenuItem('System', 'Stats', 'processes', 'open-processes')
vd.addMenuItem('System', 'Stats', 'resource limits', 'open-rlimits')



import time

from visidata import vd, VisiData, BaseSheet, Sheet, TextSheet, PyobjSheet
from visidata import ItemColumn, Column, vlen, date, asyncsingle, ENTER, AttrDict

vd.option('zulip_batch_size', -100, 'number of messages to fetch per call (<0 to fetch before anchor)')
vd.option('zulip_anchor', 1000000000, 'message id to start fetching from')
vd.option('zulip_delay_s', 0.00001, 'seconds to wait between calls (0 to stop after first)')
vd.option('zulip_api_key', '', 'Zulip API key')
vd.option('zulip_email', '', 'Email for use with Zulip API key')


@VisiData.api
def open_zulip(vd, p):
    vd.importExternal('zulip')
    import zulip

    if not vd.options.zulip_api_key:
        vd.warning('zulip_api_key must be set first')
        vd.status('Enter your login email and Zulip API key (see _https://zulip.com/api/api-keys_).')
        email = vd.input(f'Login email for {p.given}: ', record=False)
        api_key = vd.input(f'Zulip API key: ', record=False)

        vd.setPersistentOptions(zulip_email=email, zulip_api_key=api_key)

    vd.z_client = zulip.Client(site=p.given, api_key=vd.options.zulip_api_key, email=vd.options.zulip_email)

    return vd.subscribedStreams


VisiData.openhttp_zulip = VisiData.open_zulip


@VisiData.api
def z_rpc(vd, r, result_field_name=None):
    if r['result'] != 'success':
        return PyobjSheet(result_field_name+'_error', source=r)
    elif result_field_name:
        return PyobjSheet(result_field_name, source=r[result_field_name])


@VisiData.lazy_property
def allStreams(vd):
    return ZulipStreamsSheet("all_streams", zulip_func='get_streams', zulip_result_key="streams", zulip_kwargs=dict(include_public=True, include_subscribed=True))


@VisiData.lazy_property
def subscribedStreams(vd):
    return ZulipStreamsSheet("subscriptions", zulip_func='get_subscriptions', zulip_result_key="subscriptions")


@VisiData.lazy_property
def allMessages(vd):
    return ZulipMessagesSheet("all_messages")


@VisiData.api
def parseColumns(vd, fieldlist):
    for cname in fieldlist:
        kwargs = {}
        while not cname[0].isalpha():
            if cname[0] == '#': kwargs['type'] = int
            elif cname[0] == '@': kwargs['type'] = date
            elif cname[0] == '-': kwargs['width'] = 0
            else: break
            cname = cname[1:]
        yield ItemColumn(cname, **kwargs)


class ZulipAPISheet(Sheet):
    zulip_func = None
    zulip_result_key = ''
    zulip_args = []
    zulip_kwargs = {}
    fields = ''

    def iterload(self):
        self.columns = []
        for c in vd.parseColumns(self.fields.split()):
            self.addColumn(c)

        zulip_func = self.zulip_func
        if isinstance(zulip_func, str): # allow later binding for startup perf
            zulip_func = getattr(vd.z_client, zulip_func)
        r = zulip_func(*self.zulip_args, **self.zulip_kwargs)
        if r['result'] != 'success':
            vd.push(PyobjSheet(self.zulip_result_key+'_error', source=r))
            return
        yield from r[self.zulip_result_key]

    def addRow(self, r, **kwargs):
        return super().addRow(AttrDict(r), **kwargs)


class ZulipStreamsSheet(ZulipAPISheet):
    guide = '''# Zulip Streams

- `Enter` to open recent messages from the stream
- `z Enter` to open list of topics from the stream
'''
    rowtype = 'streams'  # rowdef: dict of stream from server
    fields = '-#stream_id name @date_created description -rendered_description -invite_only -is_web_public -stream_post_policy -history_public_to_subscribers -#first_message_id -#message_retention_days -is_announcement_only'

    def openRow(self, r):
        return ZulipMessagesSheet(r.name, filters=dict(stream=r.name))

    def openCell(self, c, r):
        return ZulipTopicsSheet(r.name+'_topics',
                             zulip_func=vd.z_client.get_stream_topics,
                             zulip_args=[r.stream_id],
                             zulip_result_key='topics')


class ZulipTopicsSheet(ZulipAPISheet):
    rowtype = 'topics'  # rowdef: dict of topic from server
    fields='name #max_id'
    def openRow(self, r):
        return ZulipMessagesSheet(f'{r.name}:{r.subject}', filters=dict(stream=r.name, topic=r.subject))


class ZulipMembersSheet(ZulipAPISheet):
    guide = '''# Zulip Members
- `Enter` to open list of messages from this member
'''
    rowtype = 'members'  # rowdef: dict of member from server
    fields = '''-#user_id full_name email timezone @date_joined -#avatar_version -is_admin -is_owner -is_guest -is_bot -#role -is_active -avatar_url -bot_type -#bot_owner_id'''
    def openRow(self, r):
        return ZulipMessagesSheet(r.display_recipient, filters=dict(stream=r.display_recipient))


class ZulipMessagesSheet(Sheet):
    guide = '''# Zulip Messages Sheet
Loads continuously starting with most recent, until all messages have been read.

- `Ctrl+C` to cancel loading.
- `Enter` to open message in word-wrapped text sheet
'''
    rowtype = 'messages'  # rowdef: dict of message from server
#    fields = ''
    columns = [
        ItemColumn('timestamp', type=date, fmtstr='%Y-%m-%d %H:%M'),
        ItemColumn('sender', 'sender_full_name'),
        ItemColumn('sender_email', width=0),
        ItemColumn('recipient', 'display_recipient'),
        ItemColumn('subject'),
        ItemColumn('content'),
        ItemColumn('client', width=0),
        ItemColumn('reactions', type=vlen),
        ItemColumn('submessages', type=vlen),
        ItemColumn('flags', width=0),
    ]
    filters={}

    @asyncsingle # kill previous thread
    def reload(self):
        self.rows = []
        narrow = list(self.filters.items())
        n = self.options.zulip_batch_size
        req = AttrDict(
            num_before = -n if n < 0 else 0,
            num_after = n if n > 0 else 0,
            anchor = self.options.zulip_anchor,
            apply_markdown = False,
            narrow = narrow)

        while True:
            r = vd.z_client.call_endpoint(url='messages', method='GET', request=req)
            if r['result'] == 'success':
                if not r['messages']: break
                for i, msg in enumerate(r['messages']):
                    self.addRow(msg, index=i)
                req['anchor'] = min(msg['id'] for msg in r['messages'])-1
                s = self.options.zulip_delay_s
                if s <= 0:
                    break
                time.sleep(s)

#        vd.status('finished loading, waiting for new messages')
#        vd.z_client.call_on_each_event(self.received_event, ['message'], narrow=narrow)

    def get_channel_name(self, r):
        recp = r['display_recipient']
        if isinstance(recp, list):  # private message
            return '[%s]' % recp[0]['full_name']
        else:
            return '%s:%s' % (recp, r['subject'])

    def update_message(self, msgid, content):
        req = {
            "message_id": msgid,
            "content": content
        }
        vd.z_rpc(vd.z_client.update_message(req))

    def openRow(self, r):
        vs = TextSheet(self.get_channel_name(r), source=[r["content"]])
        vs.options.wrap = True
        return vs

    def received_event(self, event):
        if event['type'] == 'message':
            self.addRow(event['message'])

    def reply_message(self, msg, row):
        recp = row['display_recipient']
        if isinstance(recp, list):
            for dest in recp:
                self.send_message(msg, row['subject'], dest['email'], 'private')
        else:
            self.send_message(msg, row['subject'], dest, 'stream')

    def send_message(self, msg, subject, dest, msgtype='stream'):
        req = {
            'type': msgtype,
            'content': msg,
            'subject': subject,
            'to': dest,
        }
        r = vd.z_client.send_message(req)

        if r['result'] != 'success':
            vd.push(PyobjSheet('send_message_result', source=r))


vd.addGlobals({
    'ZulipMembersSheet': ZulipMembersSheet,
    'ZulipStreamsSheet': ZulipStreamsSheet,
    'ZulipAPISheet': ZulipAPISheet,
    'ZulipMessagesSheet': ZulipMessagesSheet,
})

ZulipAPISheet.addCommand('', 'open-zulip-profile', 'vd.push(PyobjSheet("profile", source=z_client.get_profile()))', 'open connected user\'s profile')
ZulipAPISheet.addCommand('', 'open-zulip-members', 'vd.push(ZulipMembersSheet("members", zulip_func=z_client.get_users, zulip_result_key="members"))', 'open list of all members')
ZulipAPISheet.addCommand('', 'open-zulip-streams', 'vd.push(vd.allStreams)', 'open list of all streams')
ZulipAPISheet.addCommand('', 'open-zulip-subs', 'vd.push(vd.subscribedStreams)', 'open list of subscribed streams')
ZulipAPISheet.addCommand('', 'open-zulip-msgs', 'vd.push(vd.allMessages)', 'open list of all messages')

ZulipMessagesSheet.addCommand('', 'reply-zulip-msg', 'reply_message(input(cursorRow["display_recipient"][1]["short_name"]+"> ", "message"), cursorRow)', 'reply to current topic')
ZulipMessagesSheet.addCommand('', 'edit-zulip-msg', 'update_message(cursorRow["id"], editCell(3, cursorRowIndex))', 'edit message content')

vd.addMenuItems('''
File > Zulip > profile > open-zulip-profile
File > Zulip > member list > open-zulip-members
File > Zulip > streams > open-zulip-streams
File > Zulip > subscriptions > open-zulip-subs
File > Zulip > messages > open-zulip-subs
File > Zulip > reply > reply-zulip-msg
File > Zulip > edit message > edit-zulip-msg
''')



import functools

from visidata import VisiData, vd
import visidata

def deprecated_alias(depver, *args, **kwargs):
    # expand this to create cmd
    # with cmd.deprecated=depver
    return visidata.BaseSheet.bindkey(*args, **kwargs)

@VisiData.api
def deprecated_warn(vd, funcname, ver, instead):
    import traceback

    msg = f'{funcname} deprecated since v{ver}'
    if instead:
        msg += f'; use {instead}'

    vd.warning(msg)

    if vd.options.debug:
        for line in reversed(traceback.extract_stack(limit=7)[:-2]):
            vd.warning(f'    {line.name} at {line.filename}:{line.lineno}')
        vd.warning(f'Deprecated call traceback (most recent last):')


def deprecated(ver, instead='', check=True):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            vd.deprecated_warn(wrapper.__name__, ver, instead)
            return func(*args, **kwargs)

        if check and hasattr(func, '_extensible_api'):
            vd.error(f"{func.__name__}: @deprecated applied in wrong order")  #2623

        return wrapper
    return decorator


def _deprecated_api(ver, instead=''):
    'Decorator to deliberately wrap non-deprecated .api functions as a deprecated global function.  Use @deprecated instead, except in deprecated.py.'
    return deprecated(ver, instead, check=False)


@VisiData.api
@deprecated('1.6', 'vd instead of vd()')
def __call__(vd):
    'Deprecated; use plain "vd"'
    return vd


@deprecated('1.6')
def copyToClipboard(value):
    vd.error("copyToClipboard no longer implemented")
    return visidata.clipboard_copy(value)


@deprecated('1.6')
def replayableOption(optname, default, helpstr):
    vd.option(optname, default, helpstr, replay=True)

@deprecated('1.6')
def SubrowColumn(*args, **kwargs):
    return visidata.SubColumnFunc(*args, **kwargs)

@deprecated('1.6')
def DeferredSetColumn(*args, **kwargs):
    return visidata.Column(*args, defer=True, **kwargs)

@deprecated('2.0')
def bindkey_override(keystrokes, longname):
    vd.bindkeys.set(keystrokes, longname)

bindkey = visidata.BaseSheet.bindkey
unbindkey = visidata.BaseSheet.unbindkey

@visidata.Sheet.api
@deprecated('2.0')
def exec_keystrokes(self, keystrokes, vdglobals=None):
    return self.execCommand(self.getCommand(keystrokes), vdglobals, keystrokes=keystrokes)

visidata.Sheet.exec_command = deprecated('2.0')(visidata.Sheet.execCommand)

@VisiData.api
@deprecated('2.0', 'def open_<filetype> instead')
def filetype(vd, ext, constructor):
    'Add constructor to handle the given file type/extension.'
    globals().setdefault('open_'+ext, lambda p,ext=ext: constructor(p.base_stem, source=p, filetype=ext))

@VisiData.global_api
@deprecated('2.0', 'Sheet(namepart1, namepart2, ...)')
def joinSheetnames(vd, *sheetnames):
    'Concatenate sheet names in a standard way'
    return visidata.options.name_joiner.join(str(x) for x in sheetnames)

@VisiData.global_api
@deprecated('2.0', 'PyobjSheet')
def load_pyobj(*names, **kwargs):
    return visidata.PyobjSheet(*names, **kwargs)

@VisiData.global_api
@deprecated('2.0', 'PyobjSheet')
def push_pyobj(name, pyobj):
    vs = visidata.PyobjSheet(name, source=pyobj)
    if vs:
        return vd.push(vs)
    else:
        vd.error("cannot push '%s' as pyobj" % type(pyobj).__name__)

@deprecated('2.1', 'vd.isNumeric instead')
def isNumeric(col):
    return vd.isNumeric(col)

visidata.addGlobals({'load_pyobj': load_pyobj, 'isNumeric': isNumeric})

# The longnames on the left are deprecated for 2.0

deprecated_alias('2.0', 'edit-cells', 'setcol-input')
deprecated_alias('2.0', 'fill-nulls', 'setcol-fill')
deprecated_alias('2.0', 'paste-cells', 'setcol-clipboard')
deprecated_alias('2.0', 'frequency-rows', 'frequency-summary')
deprecated_alias('2.0', 'dup-cell', 'dive-cell')
deprecated_alias('2.0', 'dup-row', 'dive-row')
deprecated_alias('2.0', 'next-search', 'search-next')
deprecated_alias('2.0', 'prev-search', 'search-prev')
deprecated_alias('2.0', 'search-prev', 'searchr-next')
deprecated_alias('2.0', 'prev-sheet', 'jump-prev')
deprecated_alias('2.0', 'prev-value', 'go-prev-value')
deprecated_alias('2.0', 'next-value', 'go-next-value')
deprecated_alias('2.0', 'prev-selected', 'go-prev-selected')
deprecated_alias('2.0', 'next-selected', 'go-next-selected')
deprecated_alias('2.0', 'prev-null', 'go-prev-null')
deprecated_alias('2.0', 'next-null', 'go-next-null')
deprecated_alias('2.0', 'page-right', 'go-right-page')
deprecated_alias('2.0', 'page-left', 'go-left-page')
deprecated_alias('2.0', 'dive-cell', 'open-cell')
deprecated_alias('2.0', 'dive-row', 'open-row')
deprecated_alias('2.0', 'add-sheet', 'open-new')
deprecated_alias('2.0', 'save-sheets-selected', 'save-selected')
deprecated_alias('2.0', 'join-sheets', 'join-selected')
deprecated_alias('2.0', 'dive-rows', 'dive-selected')

# v2.3
deprecated_alias('2.3', 'show-aggregate', 'memo-aggregate')
#theme('use_default_colors', True, 'curses use default terminal colors')
#option('expand_col_scanrows', 1000, 'number of rows to check when expanding columns (0 = all)')

# 2.6

def clean_name(s):
    return visidata.vd.cleanName(s)

def maybe_clean(s, vs):
    if (vs or visidata.vd).options.clean_names:
        s = visidata.vd.cleanName(s)
    return s

def load_tsv(fn):
    vs = open_tsv(Path(fn))
    yield from vs.iterload()

# NOTE: you cannot use deprecated() with nonfuncs

cancelThread = _deprecated_api('2.6', 'vd.cancelThread')(vd.cancelThread)
status = _deprecated_api('2.6', 'vd.status')(vd.status)
warning = _deprecated_api('2.6', 'vd.warning')(vd.warning)
error = _deprecated_api('2.6', 'vd.error')(vd.error)
debug = _deprecated_api('2.6', 'vd.debug')(vd.debug)
fail = _deprecated_api('2.6', 'vd.fail')(vd.fail)

option = theme = vd.option # deprecated('2.6', 'vd.option')(vd.option)
jointypes = vd.jointypes # deprecated('2.6', 'vd.jointypes')(vd.jointypes)
confirm = _deprecated_api('2.6', 'vd.confirm')(vd.confirm)
launchExternalEditor = _deprecated_api('2.6', 'vd.launchExternalEditor')(vd.launchExternalEditor)
launchEditor = _deprecated_api('2.6', 'vd.launchEditor')(vd.launchEditor)
exceptionCaught = _deprecated_api('2.6', 'vd.exceptionCaught')(vd.exceptionCaught)
openSource = _deprecated_api('2.6', 'vd.openSource')(vd.openSource)
globalCommand = visidata.BaseSheet.addCommand
visidata.Sheet.StaticColumn = _deprecated_api('2.11', 'Sheet.freeze_col')(visidata.Sheet.freeze_col)
#visidata.Path.open_text = deprecated('3.0', 'visidata.Path.open')(visidata.Path.open)  # undeprecated in 3.1

vd.sysclip_value = _deprecated_api('3.0', 'vd.sysclipValue')(vd.sysclipValue)

def itemsetter(i):
    def g(obj, v):
        obj[i] = v
    return g


vd.optalias('force_valid_colnames', 'clean_names')
vd.optalias('dir_recurse', 'dir_depth', 100000)
vd.optalias('confirm_overwrite', 'overwrite', 'confirm')
vd.optalias('show_graph_labels', 'disp_graph_labels')
vd.optalias('zoom_incr', 'disp_zoom_incr')

deprecated_alias('3.0', 'visibility-sheet', 'toggle-multiline')
deprecated_alias('3.0', 'visibility-col', 'toggle-multiline')

def clean_to_id(s):
    return visidata.vd.cleanName(s)

@deprecated('3.0', 'use try/finally')
class OnExit:
    '"with OnExit(func, ...):" calls func(...) when the context is exited'
    def __init__(self, func, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        try:
            self.func(*self.args, **self.kwargs)
        except Exception as e:
            vd.exceptionCaught(e)

deprecated_alias('3.0', 'open-inputs', 'open-input-history')

#vd.option('plugins_url', 'https://visidata.org/plugins/plugins.jsonl', 'source of plugins sheet')

@visidata.VisiData.api
def inputRegexSubstOld(vd, prompt):
    'Input regex transform via oneliner (separated with `/`).  Return parsed transformer as dict(before=, after=).'
    rex = vd.inputRegex(prompt, type='regex-subst')
    before, after = vd.parse_sed_transform(rex)
    return dict(before=before, after=after)


visidata.Sheet.addCommand('', 'addcol-subst', 'addColumnAtCursor(Column(cursorCol.name + "_re", getter=regexTransform(cursorCol, **inputRegexSubstOld("transform column by regex: "))))', 'add column derived from current column, replacing regex with subst (may include \1 backrefs)', deprecated='3.0')
visidata.Sheet.addCommand('', 'setcol-subst', 'setValuesFromRegex([cursorCol], someSelectedRows, **inputRegexSubstOld("transform column by regex: "))', 'regex/subst - modify selected rows in current column, replacing regex with subst, (may include backreferences \\1 etc)', deprecated='3.0')
visidata.Sheet.addCommand('', 'setcol-subst-all', 'setValuesFromRegex(visibleCols, someSelectedRows, **inputRegexSubstOld(f"transform {nVisibleCols} columns by regex: "))', 'modify selected rows in all visible columns, replacing regex with subst (may include \\1 backrefs)', deprecated='3.0')

visidata.Sheet.addCommand('', 'split-col', 'addRegexColumns(makeRegexSplitter, cursorCol, inputRegex("split regex: ", type="regex-split"))', 'Add new columns from regex split', deprecated='3.0')
visidata.Sheet.addCommand('', 'capture-col', 'addRegexColumns(makeRegexMatcher, cursorCol, inputRegex("capture regex: ", type="regex-capture"))', 'add new column from capture groups of regex; requires example row', deprecated='3.0')

#vd.option('cmdlog_histfile', '', 'file to autorecord each cmdlog action to', sheettype=None)
#BaseSheet.bindkey('KEY_BACKSPACE', 'menu-help')

@visidata.Column.api
@deprecated('3.0', 'vd.callNoExceptions(col.setValue, row, value)')
def setValueSafe(self, row, value):
    'setValue and ignore exceptions.'
    return vd.callNoExceptions(self.setValue, row, value)

@visidata.BaseSheet.api
@deprecated('3.0', 'vd.callNoExceptions(sheet.checkCursor)')
def checkCursorNoExceptions(sheet):
    return vd.callNoExceptions(sheet.checkCursor)

@VisiData.api
@deprecated('3.1', 'vd.memoValue(name, value, displayvalue)')
def memo(vd, name, col, row):
    return vd.memoValue(name, col.getTypedValue(row), col.getDisplayValue(row))

deprecated_alias('3.1', 'view-cell', 'pyobj-cell')

vd.optalias('textwrap_cells', 'disp_wrap_max_lines', 3) # wordwrap text for multiline rows

@visidata.TableSheet.api
@deprecated('3.1', 'sheet.rowname(row)')
def keystr(sheet, row):
    return sheet.rowname(row)

vd.optalias('color_refline', 'color_graph_refline') # color_refline was used in v3.1 by mistake

@visidata.TableSheet.api
@deprecated('3.2', '[self.unsetKeys([c]) if c.keycol else self.setKeys([c]) for c in cols]')
def toggleKeys(self, cols):
    for col in cols:
        if col.keycol:
            self.unsetKeys([col])
        else:
            self.setKeys([col])

vd.optalias('disp_pixel_random', 'disp_graph_pixel_random')  #2661

vd.addGlobals(deprecated_warn=deprecated_warn)

# v3.3

#vd.option('disp_expert', 'max level of options and columns to include')
#vd.option('disp_help', '')
